from typing import Optional

import torch

from tensordict.tensordict import TensorDict
from torchrl.data import Bounded, Composite, Unbounded

from rl4co.envs.eda.dpp.env import DPPEnv
from rl4co.utils.pylogger import get_pylogger

from .generator import MDPPGenerator

log = get_pylogger(__name__)


class MDPPEnv(DPPEnv):
    """Multiple decap placement problem (mDPP) environment
    This is a modified version of the DPP environment where we allow multiple probing ports

    Observations:
        - locations of the probing ports and keepout regions
        - current decap placement
        - remaining decaps

    Constraints:
        - decaps cannot be placed at the probing ports or keepout regions
        - the number of decaps is limited

    Finish Condition:
        - the number of decaps exceeds the limit

    Reward:
        - the impedance suppression at the probing ports

    Args:
        generator: DPPGenerator instance as the data generator
        generator_params: parameters for the generator
        reward_type: reward type, either minmax or meansum
            - minmax: min of the max of the decap scores
            - meansum: mean of the sum of the decap scores

    Note:
        The minmax is more challenging as it requires to find the best decap location
        for the worst case
    """

    name = "mdpp"

    def __init__(
        self,
        generator: MDPPGenerator = None,
        generator_params: dict = {},
        reward_type: str = "minmax",
        **kwargs,
    ):
        super().__init__(**kwargs)
        if generator is None:
            generator = MDPPGenerator(**generator_params)
        self.generator = generator

        assert reward_type in [
            "minmax",
            "meansum",
        ], "reward_type must be minmax or meansum"
        self.reward_type = reward_type

        self._make_spec(self.generator)

    def _step(self, td: TensorDict) -> TensorDict:
        # Step function is the same as DPPEnv, only masking changes
        return super()._step(td)

    def _reset(self, td: Optional[TensorDict] = None, batch_size=None) -> TensorDict:
        # Reset function is the same as DPPEnv, only masking changes due to probes
        td_reset = super()._reset(td, batch_size=batch_size)

        # Action mask is 0 if both action_mask (e.g. keepout) and probe are 0
        action_mask = torch.logical_and(td_reset["action_mask"], ~td_reset["probe"])
        # Keepout regions are the inverse of action_mask
        td_reset.update(
            {
                "keepout": ~td_reset["action_mask"],
                "action_mask": action_mask,
            }
        )
        return td_reset

    def _make_spec(self, generator: MDPPGenerator):
        self.observation_spec = Composite(
            locs=Bounded(
                low=generator.min_loc,
                high=generator.max_loc,
                shape=(generator.size**2, 2),
                dtype=torch.float32,
            ),
            probe=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            keepout=Unbounded(
                shape=(generator.size**2),
                dtype=torch.bool,
            ),
            i=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            action_mask=Unbounded(
                shape=(generator.size**2),
                dtype=torch.bool,
            ),
            shape=(),
        )
        self.action_spec = Bounded(
            shape=(1,),
            dtype=torch.int64,
            low=0,
            high=generator.size**2,
        )
        self.reward_spec = Unbounded(shape=(1,))
        self.done_spec = Unbounded(shape=(1,), dtype=torch.bool)

    def _get_reward(self, td, actions):
        """We call the reward function with the final sequence of actions to get the reward
        Calling per-step would be very time consuming due to decap simulation
        """
        # We do the operation in a batch
        if len(td.batch_size) == 0:
            td = td.unsqueeze(0)
            actions = actions.unsqueeze(0)

        # Reward calculation is expensive since we need to run decap simulation (not vectorizable)
        reward = torch.stack(
            [
                self._single_env_reward(td_single, action)
                for td_single, action in zip(td, actions)
            ]
        )
        return reward

    @staticmethod
    def check_solution_validity(td: TensorDict, actions: torch.Tensor):
        assert True, "Not implemented"

    def _single_env_reward(self, td, actions):
        """Get reward for single environment. We"""

        list_probe = torch.nonzero(td["probe"]).squeeze()
        scores = torch.zeros_like(list_probe, dtype=torch.float32)
        for i, probe in enumerate(list_probe):
            # Get the decap scores for the probe location
            scores[i] = self._decap_simulator(probe, actions)
        # If minmax, return min of max decap scores else mean
        return scores.min() if self.reward_type == "minmax" else scores.mean()



from typing import Callable, Literal, Optional, Union

import numpy as np
import torch

from einops import rearrange
from tensordict import TensorDict

from rl4co.envs import RL4COEnvBase, get_env
from rl4co.models.common.constructive.nonautoregressive import (
    NonAutoregressiveDecoder,
    NonAutoregressiveEncoder,
    NonAutoregressivePolicy,
)
from rl4co.models.nn.env_embeddings.edge import VRPPolarEdgeEmbedding
from rl4co.models.nn.env_embeddings.init import VRPPolarInitEmbedding
from rl4co.models.rl.common.base import RL4COLitModule
from rl4co.models.zoo.glop.adapter import adapter_map
from rl4co.models.zoo.nargnn.encoder import NARGNNEncoder
from rl4co.utils.ops import batchify, select_start_nodes_by_distance
from rl4co.utils.pylogger import get_pylogger

try:
    import random_insertion as insertion
except ImportError:
    insertion = None

log = get_pylogger(__name__)

SubProblemSolverType = Union[
    Literal["insertion"],
    RL4COLitModule,
    tuple[RL4COLitModule, dict],
    Callable[[torch.Tensor], torch.Tensor],
]


class GLOPPolicy(NonAutoregressivePolicy):
    """Implements GLOP policy based on :class:`NonAutoregressivePolicy`. Introduced by Ye et al. (2023): https://arxiv.org/abs/2312.08224.
    This policy combines global partitioning with local optimization to solve routing problems.

    Args:
        encoder: :class:`NonAutoregressiveEncoder` instance for encoding problem states
        decoder: :class:`NonAutoregressiveDecoder` instance for decoding solutions
        env_name: Name of the environment (default: "cvrp")
        n_samples: Number of samples per instance for multistart decoding (default: 10)
        temperature: Temperature parameter for sampling (default: 1.0)
        subprob_adapter_class: Class for adapting global partitions to local subproblems
        subprob_adapter_kwargs: Additional arguments for subproblem adapter initialization
        subprob_solver: Solver for local subproblems (default: "insertion")
        **encoder_kwargs: Additional arguments for encoder initialization
    """

    def __init__(
        self,
        encoder: Optional[NonAutoregressiveEncoder] = None,
        decoder: Optional[NonAutoregressiveDecoder] = None,
        env_name: str = "cvrp",
        n_samples: int = 10,
        temperature: float = 1.0,
        subprob_adapter_class=None,
        subprob_adapter_kwargs: dict = {},
        subprob_solver: SubProblemSolverType = "insertion",
        **encoder_kwargs,
    ):

        if subprob_adapter_class is None:
            assert (
                env_name in adapter_map
            ), f"{env_name} is not supported by {self.__class__.__name__} yet"
            subprob_adapter_class = adapter_map.get(env_name)
            assert (
                subprob_adapter_class is not None
            ), "Can not import adapter module. Please check if `numba` is installed."

        if encoder is None:
            encoder_kwargs.setdefault("embed_dim", 64)
            if env_name.startswith("cvrp"):
                # Use Polar coordinate embeddings for CVRP as in Ye et al. (2023)
                embed_dim = encoder_kwargs.get("embed_dim", 64)
                if "init_embedding" not in encoder_kwargs:
                    encoder_kwargs["init_embedding"] = VRPPolarInitEmbedding(
                        embed_dim, attach_cartesian_coords=True
                    )
                if "edge_embedding" not in encoder_kwargs:
                    encoder_kwargs["edge_embedding"] = VRPPolarEdgeEmbedding(embed_dim)
            encoder = NARGNNEncoder(env_name=env_name, **encoder_kwargs)
        if decoder is None:
            decoder = NonAutoregressiveDecoder()

        super().__init__(
            encoder=encoder,
            decoder=decoder,
            env_name=env_name,
            temperature=temperature,
            train_decode_type="multistart_sampling",
            val_decode_type="multistart_greedy",
            test_decode_type="multistart_greedy",
        )

        self.n_samples = n_samples
        self.subprob_solver: SubProblemSolverType = subprob_solver
        self.subprob_adapter_class = subprob_adapter_class
        self.subprob_adapter_kwargs = subprob_adapter_kwargs
        self.subprob_adapter_kwargs.setdefault("subprob_batch_size", 2000)

    def forward(
        self,
        td: TensorDict,
        env: Optional[Union[RL4COEnvBase, str]] = None,
        phase: Literal["train", "val", "test"] = "test",
        calc_reward: bool = True,
        return_actions: bool = False,
        return_entropy: bool = False,
        return_init_embeds: bool = False,
        return_sum_log_likelihood: bool = False,
        subprob_solver: Optional[SubProblemSolverType] = None,
        **decoding_kwargs,
    ) -> dict:
        """Forward pass of GLOP.

        Args:
            td: TensorDict containing problem state
            env: Environment instance or name (default: None)
            phase: Current phase (train/val/test) (default: "test")
            calc_reward: Whether to calculate reward (default: True)
            return_actions: Whether to return actions (default: False)
            return_entropy: Whether to return entropy (default: False)
            return_init_embeds: Whether to return initial embeddings (default: False)
            return_sum_log_likelihood: Whether to return sum of log likelihoods (default: False)
            subprob_solver: Overriding the solver for local subproblems (default: None)
            **decoding_kwargs: Additional decoding arguments
        """

        if (
            env is None
            and self.env_name.startswith("cvrp")
            or isinstance(env, str)
            and env.startswith("cvrp")
            or isinstance(env, RL4COEnvBase)
            and env.name.startswith("cvrp")
        ):
            decoding_kwargs.setdefault(
                "select_start_nodes_fn", select_start_nodes_by_distance
            )

        par_out = super().forward(
            td=td,
            env=env,
            phase=phase,
            calc_reward=False,  # We don't need the partition reward
            return_actions=True,  # Used for partition
            return_entropy=return_entropy,
            return_init_embeds=return_init_embeds,
            return_sum_log_likelihood=return_sum_log_likelihood or phase == "train",
            num_starts=self.n_samples,
            **decoding_kwargs,
        )

        # local policy
        par_actions = par_out["actions"]
        local_policy_out = self.local_policy(
            td,
            par_actions,
            subprob_solver=self._get_subprob_solver(subprob_solver),
        )
        actions = local_policy_out["actions"]

        # Construct final output
        out = par_out

        if return_actions:
            out["actions"] = actions
        else:
            del out["actions"]

        if calc_reward:
            if isinstance(env, str) or env is None:
                env_name = self.env_name if env is None else env
                log.info(
                    f"Instantiated environment not provided; instantiating {env_name}"
                )
                env = get_env(env_name)
            td_repeated = batchify(td, self.n_samples)
            reward = env.get_reward(td_repeated, actions)
            out["reward"] = reward.detach()
        return out

    @torch.no_grad()
    def local_policy(
        self,
        td: TensorDict,
        actions: torch.Tensor,
        subprob_solver: Callable[[torch.Tensor], torch.Tensor],
    ):
        """Partition and apply local optimization policy.

        Args:
            td: TensorDict containing problem state
            actions: Global partition actions from the main policy
            subprob_solver: Function to solve local subproblems
        """

        assert self.subprob_adapter_class is not None
        actions = rearrange(actions, "(n b) ... -> (b n) ...", n=self.n_samples)

        adapter = self.subprob_adapter_class(td, actions, **self.subprob_adapter_kwargs)
        for mapping in adapter.get_batched_subprobs():
            subprob_actions = subprob_solver(mapping.subprob_coordinates)
            adapter.update_actions(mapping, subprob_actions)

        actions_revised = adapter.get_actions().to(td.device)
        actions_revised = rearrange(
            actions_revised, "(b n) ... -> (n b) ...", n=self.n_samples
        )
        return dict(actions=actions_revised)

    def _get_subprob_solver(
        self, solver: Optional[SubProblemSolverType]
    ) -> Callable[[torch.Tensor], torch.Tensor]:
        solver = self.subprob_solver if solver is None else solver
        env_name = self.subprob_adapter_class.subproblem_env_name

        if isinstance(solver, str):
            if solver == "insertion":
                if insertion is None:
                    raise ImportError(
                        "Module `random-insertion` not found. "
                        "Please try installing the module with pip or use alternate sub-problem solvers."
                    )
                if env_name == "tsp":
                    subprob_solver = self._insertion_solver_wrapper(
                        insertion.tsp_random_insertion_parallel
                    )
                elif env_name == "shpp":
                    subprob_solver = self._insertion_solver_wrapper(
                        insertion.shpp_random_insertion_parallel
                    )
                else:
                    raise NotImplementedError(f"{env_name} is not supported by insertion")
            else:
                raise ValueError(f"Unexpected sub-problem solver value '{solver}'")

        elif isinstance(solver, (RL4COLitModule, tuple)):
            env = get_env(env_name)
            if isinstance(solver, tuple):
                solver, kwargs = solver
                assert isinstance(solver, RL4COLitModule)
            else:
                kwargs = {}

            def solver_function(coordinates: torch.Tensor) -> torch.Tensor:
                td = TensorDict({"locs": coordinates}, batch_size=coordinates.shape[0])
                td = env.reset(td=td).to(solver.device)
                results = solver(td=td, env=env, **kwargs)
                return results["actions"].to(coordinates.device)

            subprob_solver = solver_function
        else:
            subprob_solver = solver

        return subprob_solver

    @staticmethod
    def _insertion_solver_wrapper(func):
        def wrapped(coords):
            results = func(coords.numpy())
            actions = torch.from_numpy(results.astype(np.int64))
            return actions

        return wrapped



import torch

from einops import rearrange, reduce
from tensordict.tensordict import TensorDict
from torchrl.data import Bounded, Composite, Unbounded

from rl4co.envs.common.base import RL4COEnvBase as EnvBase
from rl4co.utils.ops import gather_by_index, sample_n_random_actions

from . import INIT_FINISH, NO_OP_ID
from .generator import FJSPFileGenerator, FJSPGenerator
from .render import render
from .utils import calc_lower_bound, get_job_ops_mapping, op_is_ready


class FJSPEnv(EnvBase):
    """Flexible Job-Shop Scheduling Problem (FJSP) environment
    At each step, the agent chooses a job-machine combination. The operation to be processed next for the selected job is
    then executed on the selected machine. The reward is 0 unless the agent scheduled all operations of all jobs.
    In that case, the reward is (-)makespan of the schedule: maximizing the reward is equivalent to minimizing the makespan.

    Observations:
        - time: current time
        - next_op: next operation per job
        - proc_times: processing time of operation-machine pairs
        - pad_mask: specifies padded operations
        - start_op_per_job: id of first operation per job
        - end_op_per_job: id of last operation per job
        - start_times: start time of operation (defaults to 0 if not scheduled)
        - finish_times: finish time of operation (defaults to INIT_FINISH if not scheduled)
        - job_ops_adj: adjacency matrix specifying job-operation affiliation
        - ops_job_map: same as above but using ids of jobs to indicate affiliation
        - ops_sequence_order: specifies the order in which operations have to be processed
        - ma_assignment: specifies which operation has been scheduled on which machine
        - busy_until: specifies until when the machine will be busy
        - num_eligible: number of machines that can process an operation
        - job_in_process: whether job is currently being processed
        - job_done: whether the job is done

    Constrains:
        the agent may not select:
        - machines that are currently busy
        - jobs that are done already
        - jobs that are currently processed
        - job-machine combinations, where the machine cannot process the next operation of the job

    Finish condition:
        - the agent has scheduled all operations of all jobs

    Reward:
        - the negative makespan of the final schedule

    Args:
        generator: FJSPGenerator instance as the data generator
        generator_params: parameters for the generator
        mask_no_ops: if True, agent may not select waiting operation (unless instance is done)
    """

    name = "fjsp"

    def __init__(
        self,
        generator: FJSPGenerator = None,
        generator_params: dict = {},
        mask_no_ops: bool = True,
        check_mask: bool = False,
        stepwise_reward: bool = False,
        **kwargs,
    ):
        super().__init__(check_solution=False, **kwargs)
        if generator is None:
            if generator_params.get("file_path", None) is not None:
                generator = FJSPFileGenerator(**generator_params)
            else:
                generator = FJSPGenerator(**generator_params)
        self.generator = generator
        self._num_mas = generator.num_mas
        self._num_jobs = generator.num_jobs
        self._n_ops_max = generator.max_ops_per_job * self.num_jobs

        self.mask_no_ops = mask_no_ops
        self.check_mask = check_mask
        self.stepwise_reward = stepwise_reward
        self._make_spec(self.generator)

    @property
    def num_mas(self):
        return self._num_mas

    @property
    def num_jobs(self):
        return self._num_jobs

    @property
    def n_ops_max(self):
        return self._n_ops_max

    def set_instance_params(self, td):
        self._num_jobs = td["start_op_per_job"].size(1)
        self._num_mas = td["proc_times"].size(1)
        self._n_ops_max = td["proc_times"].size(2)

    def _decode_graph_structure(self, td: TensorDict):
        batch_size = td.batch_size
        start_op_per_job = td["start_op_per_job"]
        end_op_per_job = td["end_op_per_job"]
        pad_mask = td["pad_mask"]
        n_ops_max = td["pad_mask"].size(-1)

        # here we will generate the operations-job mapping:
        ops_job_map, ops_job_bin_map = get_job_ops_mapping(
            start_op_per_job, end_op_per_job, n_ops_max
        )

        # mask invalid edges (caused by padding)
        ops_job_bin_map[pad_mask.unsqueeze(1).expand_as(ops_job_bin_map)] = 0

        # generate for each batch a sequence specifying the position of all operations in their respective jobs,
        # e.g. [0,1,0,0,1,2,0,1,2,3,0,0] for jops with n_ops=[2,1,3,4,1,1]
        # (bs, max_ops)
        ops_seq_order = torch.sum(
            ops_job_bin_map * (ops_job_bin_map.cumsum(2) - 1), dim=1
        )

        # predecessor and successor adjacency matrices
        pred = torch.diag_embed(torch.ones(n_ops_max - 1), offset=-1)[None].expand(
            *batch_size, -1, -1
        )
        # the start of the sequence (of each job) does not have a predecessor, therefore we can
        # mask all first ops of a job in the predecessor matrix
        pred = pred * ops_seq_order.gt(0).unsqueeze(-1).expand_as(pred).to(pred)
        succ = torch.diag_embed(torch.ones(n_ops_max - 1), offset=1)[None].expand(
            *batch_size, -1, -1
        )
        # apply the same logic as above to mask the last op of a job, which does not have a successor. The last job of a job
        # always comes before the 1st op of the next job, therefore performing a left shift of the ops seq tensor here
        succ = succ * torch.cat(
            (ops_seq_order[:, 1:], ops_seq_order.new_full((*batch_size, 1), 0)), dim=1
        ).gt(0).to(succ).unsqueeze(-1).expand_as(succ)

        # adjacency matrix = predecessors, successors and self loops
        # (bs, max_ops, max_ops, 2)
        ops_adj = torch.stack((pred, succ), dim=3)

        td = td.update(
            {
                "ops_adj": ops_adj,
                "job_ops_adj": ops_job_bin_map,
                "ops_job_map": ops_job_map,
                # "op_spatial_enc": ops_spatial_enc,
                "ops_sequence_order": ops_seq_order,
            }
        )

        return td, n_ops_max

    def _reset(self, td: TensorDict = None, batch_size=None) -> TensorDict:
        self.set_instance_params(td)

        td_reset = td.clone()

        td_reset, n_ops_max = self._decode_graph_structure(td_reset)

        # schedule
        start_op_per_job = td_reset["start_op_per_job"]
        start_times = torch.zeros((*batch_size, n_ops_max))
        finish_times = torch.full((*batch_size, n_ops_max), INIT_FINISH)
        ma_assignment = torch.zeros((*batch_size, self.num_mas, n_ops_max))

        # reset feature space
        busy_until = torch.zeros((*batch_size, self.num_mas))
        # (bs, ma, ops)
        ops_ma_adj = (td_reset["proc_times"] > 0).to(torch.float32)
        # (bs, ops)
        num_eligible = torch.sum(ops_ma_adj, dim=1)

        td_reset = td_reset.update(
            {
                "start_times": start_times,
                "finish_times": finish_times,
                "ma_assignment": ma_assignment,
                "busy_until": busy_until,
                "num_eligible": num_eligible,
                "next_op": start_op_per_job.clone().to(torch.int64),
                "ops_ma_adj": ops_ma_adj,
                "op_scheduled": torch.full((*batch_size, n_ops_max), False),
                "job_in_process": torch.full((*batch_size, self.num_jobs), False),
                "reward": torch.zeros((*batch_size,), dtype=torch.float32),
                "time": torch.zeros((*batch_size,)),
                "job_done": torch.full((*batch_size, self.num_jobs), False),
                "done": torch.full((*batch_size, 1), False),
            },
        )

        td_reset.set("action_mask", self.get_action_mask(td_reset))
        # add additional features to tensordict
        td_reset["lbs"] = calc_lower_bound(td_reset)
        td_reset = self._get_features(td_reset)

        return td_reset

    def _get_job_machine_availability(self, td: TensorDict):
        batch_size = td.size(0)

        # (bs, jobs, machines)
        action_mask = torch.full((batch_size, self.num_jobs, self.num_mas), False).to(
            td.device
        )

        # mask jobs that are done already
        action_mask.add_(td["job_done"].unsqueeze(2))
        # as well as jobs that are currently processed
        action_mask.add_(td["job_in_process"].unsqueeze(2))

        # mask machines that are currently busy
        action_mask.add_(td["busy_until"].gt(td["time"].unsqueeze(1)).unsqueeze(1))

        # exclude job-machine combinations, where the machine cannot process the next op of the job
        next_ops_proc_times = gather_by_index(
            td["proc_times"], td["next_op"].unsqueeze(1), dim=2, squeeze=False
        ).transpose(1, 2)
        action_mask.add_(next_ops_proc_times == 0)
        return action_mask

    def get_action_mask(self, td: TensorDict) -> torch.Tensor:
        # 1 indicates machine or job is unavailable at current time step
        action_mask = self._get_job_machine_availability(td)
        if self.mask_no_ops:
            # masking is only allowed if instance is finished
            no_op_mask = td["done"]
        else:
            # if no job is currently processed and instance is not finished yet, waiting is not allowed
            no_op_mask = (
                td["job_in_process"].any(1, keepdims=True) & (~td["done"])
            ) | td["done"]
        # flatten action mask to correspond with logit shape
        action_mask = rearrange(action_mask, "bs j m -> bs (j m)")
        # NOTE: 1 means feasible action, 0 means infeasible action
        mask = torch.cat((no_op_mask, ~action_mask), dim=1)

        return mask

    def _translate_action(self, td):
        """This function translates an action into a machine, job tuple."""
        selected_job = td["action"] // self.num_mas
        selected_op = td["next_op"].gather(1, selected_job[:, None]).squeeze(1)
        selected_machine = td["action"] % self.num_mas
        return selected_job, selected_op, selected_machine

    def _step(self, td: TensorDict):
        # cloning required to avoid inplace operation which avoids gradient backtracking
        td = td.clone()
        td["action"].subtract_(1)
        # (bs)
        dones = td["done"].squeeze(1)
        # specify which batch instances require which operation
        no_op = td["action"].eq(NO_OP_ID)
        no_op = no_op & ~dones
        req_op = ~no_op & ~dones

        # transition to next time for no op instances
        if no_op.any():
            td, dones = self._transit_to_next_time(no_op, td)

        # select only instances that perform a scheduling action
        td_op = td.masked_select(req_op)

        td_op = self._make_step(td_op)
        # update the tensordict
        td[req_op] = td_op

        # action mask
        td.set("action_mask", self.get_action_mask(td))

        step_complete = self._check_step_complete(td, dones)
        while step_complete.any():
            td, dones = self._transit_to_next_time(step_complete, td)
            td.set("action_mask", self.get_action_mask(td))
            step_complete = self._check_step_complete(td, dones)
        if self.check_mask:
            assert reduce(td["action_mask"], "bs ... -> bs", "any").all()

        if self.stepwise_reward:
            # if we require a stepwise reward, the change in the calculated lower bounds could serve as such
            lbs = calc_lower_bound(td)
            td["reward"] = -(lbs.max(1).values - td["lbs"].max(1).values)
            td["lbs"] = lbs
        else:
            td["lbs"] = calc_lower_bound(td)

        # add additional features to tensordict
        td = self._get_features(td)

        return td

    def _get_features(self, td):
        # after we have transitioned to a next time step, we determine which operations are ready
        td["is_ready"] = op_is_ready(td)
        # td["lbs"] = calc_lower_bound(td)

        return td

    @staticmethod
    def _check_step_complete(td, dones):
        """check whether there a feasible actions left to be taken during the current
        time step. If this is not the case (and the instance is not done),
        we need to adance the timer of the repsective instance
        """
        return ~reduce(td["action_mask"], "bs ... -> bs", "any") & ~dones

    def _make_step(self, td: TensorDict) -> TensorDict:
        """
        Environment transition function
        """

        batch_idx = torch.arange(td.size(0))

        # 3*(#req_op)
        selected_job, selected_op, selected_machine = self._translate_action(td)

        # mark job as being processed
        td["job_in_process"][batch_idx, selected_job] = 1

        # mark op as schedules
        td["op_scheduled"][batch_idx, selected_op] = True

        # update machine state
        proc_time_of_action = td["proc_times"][batch_idx, selected_machine, selected_op]
        # we may not select a machine that is busy
        assert torch.all(td["busy_until"][batch_idx, selected_machine] <= td["time"])

        # update schedule
        td["start_times"][batch_idx, selected_op] = td["time"]
        td["finish_times"][batch_idx, selected_op] = td["time"] + proc_time_of_action
        td["ma_assignment"][batch_idx, selected_machine, selected_op] = 1
        # update the state of the selected machine
        td["busy_until"][batch_idx, selected_machine] = td["time"] + proc_time_of_action
        # update adjacency matrices (remove edges)
        td["proc_times"] = td["proc_times"].scatter(
            2,
            selected_op[:, None, None].expand(-1, self.num_mas, 1),
            torch.zeros_like(td["proc_times"]),
        )
        td["ops_ma_adj"] = td["proc_times"].contiguous().gt(0).to(torch.float32)
        td["num_eligible"] = torch.sum(td["ops_ma_adj"], dim=1)
        # update the positions of an operation in the job (subtract 1 from each operation of the selected job)
        td["ops_sequence_order"] = (
            td["ops_sequence_order"] - gather_by_index(td["job_ops_adj"], selected_job, 1)
        ).clip(0)
        # some checks
        # assert torch.allclose(
        #     td["proc_times"].sum(1).gt(0).sum(1),  # num ops with eligible machine
        #     (~(td["op_scheduled"] + td["pad_mask"])).sum(1),  # num unscheduled ops
        # )

        return td

    def _transit_to_next_time(self, step_complete, td: TensorDict) -> TensorDict:
        """
        Transit to the next time
        """

        # we need a transition to a next time step if either
        # 1.) all machines are busy
        # 2.) all operations are already currently in process (can only happen if num_jobs < num_machines)
        # 3.) idle machines can not process any of the not yet scheduled operations
        # 4.) no_op is choosen
        available_time_ma = td["busy_until"]
        end_op_per_job = td["end_op_per_job"]
        # we want to transition to the next time step where a machine becomes idle again. This time step must be
        # in the future, therefore we mask all machine idle times lying in the past / present
        available_time = (
            torch.where(
                available_time_ma > td["time"][:, None], available_time_ma, torch.inf
            )
            .min(1)
            .values
        )

        assert not torch.any(available_time[step_complete].isinf())
        td["time"] = torch.where(step_complete, available_time, td["time"])

        # this may only be set when the operation is finished, not when it is scheduled
        # operation of job is finished, set next operation and flag job as being idle
        curr_ops_end = td["finish_times"].gather(1, td["next_op"])
        op_finished = td["job_in_process"] & (curr_ops_end <= td["time"][:, None])
        # check whether a job is finished, which is the case when the last operation of the job is finished
        job_finished = op_finished & (td["next_op"] == end_op_per_job)
        # determine the next operation for a job that is not done, but whose latest operation is finished
        td["next_op"] = torch.where(
            op_finished & ~job_finished,
            td["next_op"] + 1,
            td["next_op"],
        )
        td["job_in_process"][op_finished] = False

        td["job_done"] = td["job_done"] + job_finished
        td["done"] = td["job_done"].all(1, keepdim=True)

        return td, td["done"].squeeze(1)

    def _get_reward(self, td, actions=None) -> TensorDict:
        if self.stepwise_reward and actions is None:
            return td["reward"]
        else:
            assert td[
                "done"
            ].all(), "Set stepwise_reward to True if you want reward prior to completion"
            return (
                -td["finish_times"].masked_fill(td["pad_mask"], -torch.inf).max(1).values
            )

    def _make_spec(self, generator: FJSPGenerator):
        self.observation_spec = Composite(
            time=Unbounded(
                shape=(1,),
                dtype=torch.int64,
            ),
            next_op=Unbounded(
                shape=(self.num_jobs,),
                dtype=torch.int64,
            ),
            proc_times=Unbounded(
                shape=(self.num_mas, self.n_ops_max),
                dtype=torch.float32,
            ),
            pad_mask=Unbounded(
                shape=(self.num_mas, self.n_ops_max),
                dtype=torch.bool,
            ),
            start_op_per_job=Unbounded(
                shape=(self.num_jobs,),
                dtype=torch.bool,
            ),
            end_op_per_job=Unbounded(
                shape=(self.num_jobs,),
                dtype=torch.bool,
            ),
            start_times=Unbounded(
                shape=(self.n_ops_max,),
                dtype=torch.int64,
            ),
            finish_times=Unbounded(
                shape=(self.n_ops_max,),
                dtype=torch.int64,
            ),
            job_ops_adj=Unbounded(
                shape=(self.num_jobs, self.n_ops_max),
                dtype=torch.int64,
            ),
            ops_job_map=Unbounded(
                shape=(self.n_ops_max),
                dtype=torch.int64,
            ),
            ops_sequence_order=Unbounded(
                shape=(self.n_ops_max),
                dtype=torch.int64,
            ),
            ma_assignment=Unbounded(
                shape=(self.num_mas, self.n_ops_max),
                dtype=torch.int64,
            ),
            busy_until=Unbounded(
                shape=(self.num_mas,),
                dtype=torch.int64,
            ),
            num_eligible=Unbounded(
                shape=(self.n_ops_max,),
                dtype=torch.int64,
            ),
            job_in_process=Unbounded(
                shape=(self.num_jobs,),
                dtype=torch.bool,
            ),
            job_done=Unbounded(
                shape=(self.num_jobs,),
                dtype=torch.bool,
            ),
            shape=(),
        )
        self.action_spec = Bounded(
            shape=(1,),
            dtype=torch.int64,
            low=-1,
            high=self.n_ops_max,
        )
        self.reward_spec = Unbounded(shape=(1,))
        self.done_spec = Unbounded(shape=(1,), dtype=torch.bool)

    @staticmethod
    def render(td, idx):
        return render(td, idx)

    def select_start_nodes(self, td: TensorDict, num_starts: int):
        return sample_n_random_actions(td, num_starts)

    def get_num_starts(self, td):
        # NOTE in the paper they use N_s = 100
        return 100

    def load_data(self, fpath, batch_size=[]):
        g = FJSPFileGenerator(fpath)
        return g(batch_size=batch_size)



from typing import Optional

import torch

from tensordict.tensordict import TensorDict
from torchrl.data import Bounded, Composite, UnboundedContinuous, UnboundedDiscrete

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.utils.ops import gather_by_index, get_tour_length
from rl4co.utils.pylogger import get_pylogger

from .generator import SHPPGenerator
from .render import render

log = get_pylogger(__name__)


class SHPPEnv(RL4COEnvBase):
    """
    Shortest Hamiltonian Path Problem (SHPP)
    SHPP is referred to the open-loop Traveling Salesman Problem (TSP) in the literature.
    The goal of the SHPP is to find the shortest Hamiltonian path in a given graph with
    given fixed starting/terminating nodes (they can be different nodes). A Hamiltonian
    path visits all other nodes exactly once. At each step, the agent chooses a city to visit.
    The reward is 0 unless the agent visits all the cities. In that case, the reward is
    (-)length of the path: maximizing the reward is equivalent to minimizing the path length.

    Observation:
        - locations of each customer
        - starting node and terminating node
        - the current location of the vehicle

    Constraints:
        - the first node is the starting node
        - the last node is the terminating node
        - each node is visited exactly once

    Finish condition:
        - the agent has visited all the customers and reached the terminating node

    Reward:
        - (minus) the length of the path

    Args:
        generator: SHPPGenerator instance as the generator
        generator_params: parameters for the generator
    """

    name = "shpp"

    def __init__(
        self,
        generator: SHPPGenerator = None,
        generator_params: dict = {},
        **kwargs,
    ):
        super().__init__(**kwargs)
        if generator is None:
            generator = SHPPGenerator(**generator_params)
        self.generator = generator
        self._make_spec(self.generator)

    @staticmethod
    def _step(td: TensorDict) -> TensorDict:
        current_node = td["action"]
        first_node = current_node if td["i"].all() == 0 else td["first_node"]

        # Set not visited to 0 (i.e., we visited the node)
        available = td["available"].scatter(
            -1, current_node.unsqueeze(-1).expand_as(td["action_mask"]), 0
        )

        # If all other nodes are visited, the terminating node will be available
        action_mask = available.clone()
        action_mask[..., -1] = ~available[..., :-1].any(dim=-1)

        # We are done there are no unvisited locations
        done = torch.sum(available, dim=-1) == 0

        # The reward is calculated outside via get_reward for efficiency, so we set it to 0 here
        reward = torch.zeros_like(done)

        td.update(
            {
                "first_node": first_node,
                "current_node": current_node,
                "i": td["i"] + 1,
                "available": available,
                "action_mask": action_mask,
                "reward": reward,
                "done": done,
            },
        )
        return td

    def _reset(self, td: Optional[TensorDict] = None, batch_size=None) -> TensorDict:
        """Note: the first node is the starting node; the last node is the terminating node"""
        device = td.device
        locs = td["locs"]

        # We do not enforce loading from self for flexibility
        num_loc = locs.shape[-2]

        # Other variables
        current_node = torch.zeros((batch_size), dtype=torch.int64, device=device)
        last_node = torch.full(
            (batch_size), num_loc - 1, dtype=torch.int64, device=device
        )
        available = torch.ones(
            (*batch_size, num_loc), dtype=torch.bool, device=device
        )  # 1 means not visited, i.e. action is allowed
        action_mask = torch.zeros((*batch_size, num_loc), dtype=torch.bool, device=device)
        action_mask[..., 0] = 1  # Only the start point is availabe at the beginning
        i = torch.zeros((*batch_size, 1), dtype=torch.int64, device=device)

        return TensorDict(
            {
                "locs": locs,
                "first_node": current_node,
                "last_node": last_node,
                "current_node": current_node,
                "i": i,
                "available": available,
                "action_mask": action_mask,
                "reward": torch.zeros((*batch_size, 1), dtype=torch.float32),
            },
            batch_size=batch_size,
        )

    def _get_reward(self, td, actions) -> TensorDict:
        # Gather locations in order of tour and return distance between them (i.e., -reward)
        locs_ordered = gather_by_index(td["locs"], actions)
        return -get_tour_length(locs_ordered)

    @staticmethod
    def check_solution_validity(td: TensorDict, actions: torch.Tensor):
        """Check that solution is valid: nodes are visited exactly once"""
        assert (
            torch.arange(actions.size(1), out=actions.data.new())
            .view(1, -1)
            .expand_as(actions)
            == actions.data.sort(1)[0]
        ).all(), "Invalid tour"

    @staticmethod
    def render(td: TensorDict, actions: torch.Tensor = None, ax=None):
        return render(td, actions, ax)

    def _make_spec(self, generator):
        """Make the observation and action specs from the parameters"""
        self.observation_spec = Composite(
            locs=Bounded(
                low=generator.min_loc,
                high=generator.max_loc,
                shape=(generator.num_loc, 2),
                dtype=torch.float32,
            ),
            first_node=UnboundedDiscrete(
                shape=(1),
                dtype=torch.int64,
            ),
            current_node=UnboundedDiscrete(
                shape=(1),
                dtype=torch.int64,
            ),
            i=UnboundedDiscrete(
                shape=(1),
                dtype=torch.int64,
            ),
            action_mask=UnboundedDiscrete(
                shape=(generator.num_loc),
                dtype=torch.bool,
            ),
            shape=(),
        )
        self.action_spec = Bounded(
            shape=(1,),
            dtype=torch.int64,
            low=0,
            high=generator.num_loc,
        )
        self.reward_spec = UnboundedContinuous(shape=(1,))
        self.done_spec = UnboundedDiscrete(shape=(1,), dtype=torch.bool)



import copy

from typing import Optional

from tensordict import TensorDict
from torch import Tensor, nn

from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class CriticNetwork(nn.Module):
    """Create a critic network given an encoder (e.g. as the one in the policy network)
    with a value head to transform the embeddings to a scalar value.

    Args:
        encoder: Encoder module to encode the input
        value_head: Value head to transform the embeddings to a scalar value
        embed_dim: Dimension of the embeddings of the value head
        hidden_dim: Dimension of the hidden layer of the value head
    """

    def __init__(
        self,
        encoder: nn.Module,
        value_head: Optional[nn.Module] = None,
        embed_dim: int = 128,
        hidden_dim: int = 512,
        customized: bool = False,
    ):
        super(CriticNetwork, self).__init__()

        self.encoder = encoder
        if value_head is None:
            # check if embed dim of encoder is different, if so, use it
            if getattr(encoder, "embed_dim", embed_dim) != embed_dim:
                log.warning(
                    f"Found encoder with different embed_dim {encoder.embed_dim} than the value head {embed_dim}. Using encoder embed_dim for value head."
                )
                embed_dim = getattr(encoder, "embed_dim", embed_dim)
            value_head = nn.Sequential(
                nn.Linear(embed_dim, hidden_dim), nn.ReLU(), nn.Linear(hidden_dim, 1)
            )
        self.value_head = value_head
        self.customized = customized

    def forward(self, x: Tensor | TensorDict, hidden=None) -> Tensor:
        """Forward pass of the critic network: encode the imput in embedding space and return the value

        Args:
            x: Input containing the environment state. Can be a Tensor or a TensorDict

        Returns:
            Value of the input state
        """
        if not self.customized:  # fir for most of costructive tasks
            h, _ = self.encoder(x)  # [batch_size, N, embed_dim] -> [batch_size, N]
            return self.value_head(h).mean(1)  # [batch_size, N] -> [batch_size]
        else:  # custimized encoder and value head with hidden input
            h = self.encoder(x)  # [batch_size, N, embed_dim] -> [batch_size, N]
            return self.value_head(h, hidden)


def create_critic_from_actor(
    policy: nn.Module, backbone: str = "encoder", **critic_kwargs
):
    # we reuse the network of the policy's backbone, such as an encoder
    encoder = getattr(policy, backbone, None)
    if encoder is None:
        raise ValueError(
            f"CriticBaseline requires a backbone in the policy network: {backbone}"
        )
    critic = CriticNetwork(copy.deepcopy(encoder), **critic_kwargs).to(
        next(policy.parameters()).device
    )
    return critic



from typing import Callable

import torch

from tensordict.tensordict import TensorDict
from torch.distributions import Uniform

from rl4co.envs.common.utils import Generator, get_sampler
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class TSPGenerator(Generator):
    """Data generator for the Travelling Salesman Problem (TSP).

    Args:
        num_loc: number of locations (customers) in the TSP
        min_loc: minimum value for the location coordinates
        max_loc: maximum value for the location coordinates
        init_sol_type: the method type used for generating initial solutions (random or greedy)
        loc_distribution: distribution for the location coordinates

    Returns:
        A TensorDict with the following keys:
            locs [batch_size, num_loc, 2]: locations of each customer
    """

    def __init__(
        self,
        num_loc: int = 20,
        min_loc: float = 0.0,
        max_loc: float = 1.0,
        init_sol_type: str = "random",
        loc_distribution: int | float | str | type | Callable = Uniform,
        **kwargs,
    ):
        self.num_loc = num_loc
        self.min_loc = min_loc
        self.max_loc = max_loc
        self.init_sol_type = init_sol_type

        # Location distribution
        if kwargs.get("loc_sampler", None) is not None:
            self.loc_sampler = kwargs["loc_sampler"]
        else:
            self.loc_sampler = get_sampler(
                "loc", loc_distribution, min_loc, max_loc, **kwargs
            )

    def _generate(self, batch_size) -> TensorDict:
        # Sample locations
        locs = self.loc_sampler.sample((*batch_size, self.num_loc, 2))

        return TensorDict(
            {
                "locs": locs,
            },
            batch_size=batch_size,
        )

    # for improvement MDP only (to be refactored by a combination of rollout and the random policy)
    def _get_initial_solutions(self, coordinates):
        batch_size = coordinates.size(0)

        if self.init_sol_type == "random":
            set = torch.rand(batch_size, self.num_loc).argsort().long()
            rec = torch.zeros(batch_size, self.num_loc).long()
            index = torch.zeros(batch_size, 1).long()

            for i in range(self.num_loc - 1):
                rec.scatter_(1, set.gather(1, index + i), set.gather(1, index + i + 1))

            rec.scatter_(1, set[:, -1].view(-1, 1), set.gather(1, index))

        elif self.init_sol_type == "greedy":
            candidates = torch.ones(batch_size, self.num_loc).bool()
            rec = torch.zeros(batch_size, self.num_loc).long()
            selected_node = torch.zeros(batch_size, 1).long()
            candidates.scatter_(1, selected_node, 0)

            for i in range(self.num_loc - 1):
                d1 = coordinates.cpu().gather(
                    1, selected_node.unsqueeze(-1).expand(batch_size, self.num_loc, 2)
                )
                d2 = coordinates.cpu()

                dists = (d1 - d2).norm(p=2, dim=2)
                dists[~candidates] = 1e5

                next_selected_node = dists.min(-1)[1].view(-1, 1)
                rec.scatter_(1, selected_node, next_selected_node)
                candidates.scatter_(1, next_selected_node, 0)
                selected_node = next_selected_node

        else:
            raise NotImplementedError()

        return rec.expand(batch_size, self.num_loc).clone()



import torch.nn as nn

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.models.rl.common.critic import CriticNetwork, create_critic_from_actor
from rl4co.models.rl.reinforce.baselines import CriticBaseline
from rl4co.models.rl.reinforce.reinforce import REINFORCE
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class A2C(REINFORCE):
    """Advantage Actor Critic (A2C) algorithm.
    A2C is a variant of REINFORCE where a baseline is provided by a critic network.
    Here we additionally support different optimizers for the actor and the critic.

    Args:
        env: Environment to use for the algorithm
        policy: Policy to use for the algorithm
        critic: Critic network to use for the algorithm
        critic_kwargs: Keyword arguments to pass to the critic network
        actor_optimizer_kwargs: Keyword arguments for the policy (=actor) optimizer
        critic_optimizer_kwargs: Keyword arguments for the critic optimizer. If None, use the same as actor_optimizer_kwargs
        **kwargs: Keyword arguments passed to the superclass
    """

    def __init__(
        self,
        env: RL4COEnvBase,
        policy: nn.Module,
        critic: CriticNetwork = None,
        critic_kwargs: dict = {},
        actor_optimizer_kwargs: dict = {"lr": 1e-4},
        critic_optimizer_kwargs: dict = None,
        **kwargs,
    ):
        if critic is None:
            log.info("Creating critic network for {}".format(env.name))
            critic = create_critic_from_actor(policy, **critic_kwargs)

        # The baseline is directly created here, so we eliminate the baseline argument
        kwargs.pop("baseline", None)

        super().__init__(env, policy, baseline=CriticBaseline(critic), **kwargs)
        self.actor_optimizer_kwargs = actor_optimizer_kwargs
        self.critic_optimizer_kwargs = (
            critic_optimizer_kwargs
            if critic_optimizer_kwargs is not None
            else actor_optimizer_kwargs
        )

    def configure_optimizers(self):
        """Configure the optimizers for the policy and the critic network (=baseline)"""
        parameters = [
            {"params": self.policy.parameters(), **self.actor_optimizer_kwargs},
        ] + [{"params": self.baseline.parameters(), **self.critic_optimizer_kwargs}]

        return super().configure_optimizers(parameters)



import os
import zipfile
from typing import Union, Callable

import torch
import numpy as np

from robust_downloader import download
from torch.distributions import Uniform
from tensordict.tensordict import TensorDict

from rl4co.data.utils import load_npz_to_tensordict
from rl4co.utils.pylogger import get_pylogger
from rl4co.envs.common.utils import get_sampler, Generator

log = get_pylogger(__name__)



class DPPGenerator(Generator):
    """Data generator for the Decap Placement Problem (DPP).

    Args:
        min_loc: Minimum location value. Defaults to 0.
        max_loc: Maximum location value. Defaults to 1.
        num_keepout_min: Minimum number of keepout regions. Defaults to 1.
        num_keepout_max: Maximum number of keepout regions. Defaults to 50.
        max_decaps: Maximum number of decaps. Defaults to 20.
        data_dir: Directory to store data. Defaults to "data/dpp/".
            This can be downloaded from this [url](https://drive.google.com/uc?id=1IEuR2v8Le-mtHWHxwTAbTOPIkkQszI95).
        chip_file: Name of the chip file. Defaults to "10x10_pkg_chip.npy".
        decap_file: Name of the decap file. Defaults to "01nF_decap.npy".
        freq_file: Name of the frequency file. Defaults to "freq_201.npy".
        url: URL to download data from. Defaults to None.
    
    Returns:
        A TensorDict with the following keys:
            locs [batch_size, num_loc, 2]: locations of each customer
            depot [batch_size, 2]: location of the depot
            demand [batch_size, num_loc]: demand of each customer
            capacity [batch_size]: capacity of the vehicle
    """
    def __init__(
        self,
        min_loc: float = 0.0,
        max_loc: float = 1.0,
        num_keepout_min: int = 1,
        num_keepout_max: int = 50,
        max_decaps: int = 20,
        data_dir: str = "data/dpp/",
        chip_file: str = "10x10_pkg_chip.npy",
        decap_file: str = "01nF_decap.npy",
        freq_file: str = "freq_201.npy",
        url: str = None,
        **unused_kwargs
    ):
        self.min_loc = min_loc
        self.max_loc = max_loc
        self.num_keepout_min = num_keepout_min
        self.num_keepout_max = num_keepout_max
        self.max_decaps = max_decaps
        self.data_dir = data_dir

        # DPP environment doen't have any other kwargs
        if len(unused_kwargs) > 0:
            log.error(f"Found {len(unused_kwargs)} unused kwargs: {unused_kwargs}")


        # Download and load the data from online dataset
        self.url = (
            "https://github.com/kaist-silab/devformer/raw/main/data/data.zip"
            if url is None
            else url
        )
        self.backup_url = (
            "https://drive.google.com/uc?id=1IEuR2v8Le-mtHWHxwTAbTOPIkkQszI95"
        )
        self._load_dpp_data(chip_file, decap_file, freq_file)

        # Check the validity of the keepout parameters
        assert (
            num_keepout_min <= num_keepout_max
        ), "num_keepout_min must be <= num_keepout_max"
        assert (
            num_keepout_max <= self.size**2
        ), "num_keepout_max must be <= size * size (total number of locations)"

    def _generate(self, batch_size) -> TensorDict:
        """
        Generate initial observations for the environment with locations, probe, and action mask
        Action_mask eliminates the keepout regions and the probe location, and is updated to eliminate placed decaps
        """
        m = n = self.size
        # if int, convert to list and make it a batch for easier generation
        batch_size = [batch_size] if isinstance(batch_size, int) else batch_size
        batched = len(batch_size) > 0
        bs = [1] if not batched else batch_size

        # Create a list of locs on a grid
        locs = torch.meshgrid(
            torch.arange(m), torch.arange(n)
        )
        locs = torch.stack(locs, dim=-1).reshape(-1, 2)
        # normalize the locations by the number of rows and columns
        locs = locs / torch.tensor([m, n], dtype=torch.float)
        locs = locs[None].expand(*bs, -1, -1)

        # Create available mask
        available = torch.ones((*bs, m * n), dtype=torch.bool)

        # Sample probe location from m*n
        probe = torch.randint(m * n, size=(*bs, 1))
        available.scatter_(1, probe, False)

        # Sample keepout locations from m*n except probe
        num_keepout = torch.randint(
            self.num_keepout_min,
            self.num_keepout_max,
            size=(*bs, 1),
        )
        keepouts = [torch.randperm(m * n)[:k] for k in num_keepout]
        for i, (a, k) in enumerate(zip(available, keepouts)):
            available[i] = a.scatter(0, k, False)

        return TensorDict(
            {
                "locs": locs if batched else locs.squeeze(0),
                "probe": probe if batched else probe.squeeze(0),
                "action_mask": available if batched else available.squeeze(0),
            },
            batch_size=batch_size,
        )

    def _load_dpp_data(self, chip_file, decap_file, freq_file):
        def _load_file(fpath):
            f = os.path.join(self.data_dir, fpath)
            if not os.path.isfile(f):
                self._download_data()
            with open(f, "rb") as f_:
                return torch.from_numpy(np.load(f_))

        self.raw_pdn = _load_file(chip_file)  # [num_freq, size^2, size^2]
        self.decap = _load_file(decap_file).to(torch.complex64)  # [num_freq, 1, 1]
        self.freq = _load_file(freq_file)  # [num_freq]
        self.size = int(np.sqrt(self.raw_pdn.shape[-1]))
        self.num_freq = self.freq.shape[0]

    def _download_data(self):
        log.info("Downloading data...")
        try:
            download(self.url, self.data_dir, "data.zip")
        except Exception:
            log.error(
                f"Download from main url {self.url} failed. Trying backup url {self.backup_url}..."
            )
            download(self.backup_url, self.data_dir, "data.zip")
        log.info("Download complete. Unzipping...")
        zipfile.ZipFile(os.path.join(self.data_dir, "data.zip"), "r").extractall(
            self.data_dir
        )
        log.info("Unzip complete. Removing zip file")
        os.remove(os.path.join(self.data_dir, "data.zip"))

    def load_data(self, fpath, batch_size=[]):
        data = load_npz_to_tensordict(fpath)
        # rename key if necessary (old dpp version)
        if "observation" in data.keys():
            data["locs"] = data.pop("observation")
        return data



from functools import partial
from typing import Optional, Type, Union

from tensordict import TensorDict
import torch

from rl4co.envs import RL4COEnvBase, get_env
from rl4co.models.common.constructive.nonautoregressive import (
    NonAutoregressiveEncoder,
    NonAutoregressivePolicy,
)
from rl4co.models.zoo.deepaco.antsystem import AntSystem
from rl4co.models.zoo.nargnn.encoder import NARGNNEncoder
from rl4co.utils.decoding import modify_logits_for_top_k_filtering, modify_logits_for_top_p_filtering
from rl4co.utils.ops import batchify, unbatchify
from rl4co.utils.utils import merge_with_defaults


class DeepACOPolicy(NonAutoregressivePolicy):
    """Implememts DeepACO policy based on :class:`NonAutoregressivePolicy`. Introduced by Ye et al. (2023): https://arxiv.org/abs/2309.14032.
    This policy uses a Non-Autoregressive Graph Neural Network to generate heatmaps,
    which are then used to run Ant Colony Optimization (ACO) to construct solutions.

    Args:
        encoder: Encoder module. Can be passed by sub-classes
        env_name: Name of the environment used to initialize embeddings
        temperature: Temperature for the softmax during decoding. Defaults to 1.0.
        aco_class: Class representing the ACO algorithm to be used. Defaults to :class:`AntSystem`.
        aco_kwargs: Additional arguments to be passed to the ACO algorithm.
        train_with_local_search: Whether to train with local search. Defaults to False.
        n_ants: Number of ants to be used in the ACO algorithm. Can be an integer or dictionary. Defaults to 20.
        n_iterations: Number of iterations to run the ACO algorithm. Can be an integer or dictionary. Defaults to `dict(train=1, val=20, test=100)`.
        encoder_kwargs: Additional arguments to be passed to the encoder.
    """

    def __init__(
        self,
        encoder: Optional[NonAutoregressiveEncoder] = None,
        env_name: str = "tsp",
        temperature: float = 1.0,
        top_p: float = 0.0,
        top_k: int = 0,
        aco_class: Optional[Type[AntSystem]] = None,
        aco_kwargs: dict = {},
        train_with_local_search: bool = False,
        n_ants: Optional[Union[int, dict]] = None,
        n_iterations: Optional[Union[int, dict]] = None,
        start_node: Optional[int] = None,
        multistart: bool = False,
        k_sparse: Optional[int] = None,
        **encoder_kwargs,
    ):
        if encoder is None:
            encoder_kwargs["k_sparse"] = k_sparse
            encoder = NARGNNEncoder(env_name=env_name, **encoder_kwargs)

        self.decode_type = "multistart_sampling" if env_name == "tsp" or multistart else "sampling"

        super().__init__(
            encoder=encoder,
            env_name=env_name,
            temperature=temperature,
            train_decode_type=self.decode_type,
            val_decode_type=self.decode_type,
            test_decode_type=self.decode_type,
        )

        self.default_decoding_kwargs = {}
        self.default_decoding_kwargs["select_best"] = False
        if k_sparse is not None:
            self.default_decoding_kwargs["top_k"] = k_sparse + (0 if env_name == "tsp" else 1)  # 1 for depot
        if "multistart" in self.decode_type:
            select_start_nodes_fn = partial(self.select_start_node_fn, start_node=start_node)
            self.default_decoding_kwargs.update(
                {"multistart": True, "select_start_nodes_fn": select_start_nodes_fn}
            )
        else:
            self.default_decoding_kwargs.update(
                {"multisample": True}
            )

        # For now, top_p and top_k are only used to filter logits (not passed to decoder)
        self.top_p = top_p
        self.top_k = top_k

        self.aco_class = AntSystem if aco_class is None else aco_class
        self.aco_kwargs = aco_kwargs
        self.train_with_local_search = train_with_local_search
        if train_with_local_search:
            assert self.aco_kwargs.get("use_local_search", False)
        self.n_ants = merge_with_defaults(n_ants, train=30, val=48, test=48)
        self.n_iterations = merge_with_defaults(n_iterations, train=1, val=5, test=10)

    @staticmethod
    def select_start_node_fn(
        td: TensorDict, env: RL4COEnvBase, num_starts: int, start_node: Optional[int] = None
    ):
        if env.name == "tsp" and start_node is not None:
            # For now, only TSP supports explicitly setting the start node
            return start_node * torch.ones(
                td.shape[0] * num_starts, dtype=torch.long, device=td.device
            )
        return torch.multinomial(td["action_mask"].float(), num_starts, replacement=True).view(-1)

    def forward(
        self,
        td_initial: TensorDict,
        env: Optional[Union[str, RL4COEnvBase]] = None,
        phase: str = "train",
        return_actions: bool = True,
        return_hidden: bool = True,
        actions=None,
        **decoding_kwargs,
    ):
        """
        Forward method. During validation and testing, the policy runs the ACO algorithm to construct solutions.
        See :class:`NonAutoregressivePolicy` for more details during the training phase.
        """
        n_ants = self.n_ants[phase]

        decoding_kwargs.update(self.default_decoding_kwargs)
        decoding_kwargs.update(
            {"num_starts": n_ants} if "multistart" in self.decode_type else {"num_samples": n_ants}
        )

        # Instantiate environment if needed
        if (phase != "train" or self.train_with_local_search) and (env is None or isinstance(env, str)):
            env_name = self.env_name if env is None else env
            env = get_env(env_name)
        else:
            assert isinstance(env, RL4COEnvBase), "env must be an instance of RL4COEnvBase"

        if phase == "train":
            #  we just use the constructive policy
            outdict = super().forward(
                td_initial,
                env,
                phase=phase,
                calc_reward=True,
                return_actions=return_actions,
                return_hidden=return_hidden,
                actions=actions,
                **decoding_kwargs,
            )

            outdict["reward"] = unbatchify(outdict["reward"], n_ants)

            if self.train_with_local_search:
                heatmap = outdict["hidden"]
                # TODO: Refactor this so that we don't need to use the aco object
                aco = self.aco_class(heatmap, n_ants=n_ants, **self.aco_kwargs)
                _, ls_reward = aco.local_search(
                    batchify(td_initial, n_ants), env, outdict["actions"], decoding_kwargs
                )
                outdict["ls_reward"] = unbatchify(ls_reward, n_ants)

            outdict["log_likelihood"] = unbatchify(outdict["log_likelihood"], n_ants)
            return outdict

        heatmap, _ = self.encoder(td_initial)
        heatmap /= self.temperature

        if self.top_k > 0:
            self.top_k = min(self.top_k, heatmap.size(-1))  # safety check
            heatmap = modify_logits_for_top_k_filtering(heatmap, self.top_k)

        if self.top_p > 0:
            assert self.top_p <= 1.0, "top-p should be in (0, 1]."
            heatmap = modify_logits_for_top_p_filtering(heatmap, self.top_p)

        aco = self.aco_class(heatmap, n_ants=n_ants, **self.aco_kwargs)
        actions, iter_rewards = aco.run(td_initial, env, self.n_iterations[phase], decoding_kwargs)

        out = {"reward": iter_rewards[self.n_iterations[phase] - 1]}
        out.update({f"reward_{i:03d}": iter_rewards[i] for i in range(self.n_iterations[phase])})
        if return_actions:
            out["actions"] = actions

        return out



import torch
import torch.nn as nn

from torch import Tensor

from rl4co.models.common.improvement.base import ImprovementDecoder
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class RDSDecoder(ImprovementDecoder):
    """
    RDS Decoder for flexible k-opt based on Ma et al. (2023)
    Given the environment state and the node embeddings (positional embeddings are discarded), compute the logits for
    selecting a k-opt exchange on basis moves (S-move, I-move, E-move) from the current solution

    Args:
        embed_dim: Embedding dimension
        num_heads: Number of attention heads
    """

    def __init__(
        self,
        embed_dim: int = 128,
    ):
        super().__init__()
        self.embed_dim = embed_dim

        self.linear_K1 = nn.Linear(self.embed_dim, self.embed_dim, bias=False)
        self.linear_K2 = nn.Linear(self.embed_dim, self.embed_dim, bias=False)
        self.linear_K3 = nn.Linear(self.embed_dim, self.embed_dim, bias=False)
        self.linear_K4 = nn.Linear(self.embed_dim, self.embed_dim, bias=False)

        self.linear_Q1 = nn.Linear(self.embed_dim, self.embed_dim, bias=False)
        self.linear_Q2 = nn.Linear(self.embed_dim, self.embed_dim, bias=False)
        self.linear_Q3 = nn.Linear(self.embed_dim, self.embed_dim, bias=False)
        self.linear_Q4 = nn.Linear(self.embed_dim, self.embed_dim, bias=False)

        self.linear_V1 = nn.Parameter(torch.Tensor(self.embed_dim))
        self.linear_V2 = nn.Parameter(torch.Tensor(self.embed_dim))

        self.rnn1 = nn.GRUCell(self.embed_dim, self.embed_dim)
        self.rnn2 = nn.GRUCell(self.embed_dim, self.embed_dim)

    def forward(self, h, q1, q2, input_q1, input_q2) -> Tensor:
        bs = h.size(0)

        # GRUs
        q1 = self.rnn1(input_q1, q1)
        q2 = self.rnn2(input_q2, q2)

        # Dual-Stream Attention
        linear_V1 = self.linear_V1.view(1, -1).expand(bs, -1)
        linear_V2 = self.linear_V2.view(1, -1).expand(bs, -1)
        result = (
            linear_V1.unsqueeze(1)
            * torch.tanh(
                self.linear_K1(h)
                + self.linear_Q1(q1).unsqueeze(1)
                + self.linear_K3(h) * self.linear_Q3(q1).unsqueeze(1)
            )
        ).sum(
            -1
        )  # \mu stream
        result += (
            linear_V2.unsqueeze(1)
            * torch.tanh(
                self.linear_K2(h)
                + self.linear_Q2(q2).unsqueeze(1)
                + self.linear_K4(h) * self.linear_Q4(q2).unsqueeze(1)
            )
        ).sum(
            -1
        )  # \lambda stream

        return result, q1, q2



import matplotlib.pyplot as plt

from tensordict.tensordict import TensorDict

from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


def render(td: TensorDict, idx: int):
    import matplotlib.patches as patches

    # TODO: fix this render function parameters
    num_machine_total = td["num_machine_total"][idx].item()
    num_job = td["num_job"][idx].item()

    job_durations = td["job_duration"][idx, :, :]
    # shape: (job, machine)
    schedule = td["schedule"][idx, :, :]
    # shape: (machine, job)

    total_machine_cnt = num_machine_total
    makespan = -td["reward"][idx].item()

    # Create figure and axes
    fig, ax = plt.subplots(figsize=(makespan / 3, 5))
    cmap = _get_cmap(num_job)

    plt.xlim(0, makespan)
    plt.ylim(0, total_machine_cnt)
    ax.invert_yaxis()

    plt.plot([0, makespan], [4, 4], "black")
    plt.plot([0, makespan], [8, 8], "black")

    for machine_idx in range(total_machine_cnt):
        duration = job_durations[:, machine_idx]
        # shape: (job)
        machine_schedule = schedule[machine_idx, :]
        # shape: (job)

        for job_idx in range(num_job):
            job_length = duration[job_idx].item()
            job_start_time = machine_schedule[job_idx].item()
            if job_start_time >= 0:
                # Create a Rectangle patch
                rect = patches.Rectangle(
                    (job_start_time, machine_idx),
                    job_length,
                    1,
                    facecolor=cmap(job_idx),
                )
                ax.add_patch(rect)

    ax.grid()
    ax.set_axisbelow(True)
    return ax


def _get_cmap(color_cnt):
    from random import shuffle

    from matplotlib.colors import CSS4_COLORS, ListedColormap

    color_list = list(CSS4_COLORS.keys())
    shuffle(color_list)
    cmap = ListedColormap(color_list, N=color_cnt)
    return cmap



from rl4co.utils.instantiators import instantiate_callbacks, instantiate_loggers
from rl4co.utils.pylogger import get_pylogger
from rl4co.utils.rich_utils import enforce_tags, print_config_tree
from rl4co.utils.trainer import RL4COTrainer
from rl4co.utils.utils import (
    extras,
    get_metric_value,
    log_hyperparameters,
    show_versions,
    task_wrapper,
)



import math

import torch
import torch.nn as nn

from tensordict import TensorDict
from torch import Tensor

from rl4co.models.common.improvement.base import ImprovementDecoder
from rl4co.models.nn.attention import MultiHeadCompat
from rl4co.models.nn.mlp import MLP
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class NodePairRemovalDecoder(ImprovementDecoder):
    """
    N2S Node-Pair Removal decoder based on Ma et al. (2022)
    Given the environment state and the node embeddings (positional embeddings are discarded), compute the logits for
    selecting a pair of pickup and delivery nodes for node pair removal from the current solution


    Args:
        embed_dim: Embedding dimension
        num_heads: Number of attention heads
    """

    def __init__(
        self,
        embed_dim: int = 128,
        num_heads: int = 4,
    ):
        super().__init__()
        self.input_dim = embed_dim
        self.n_heads = num_heads
        self.hidden_dim = embed_dim

        assert embed_dim % num_heads == 0

        self.W_Q = nn.Parameter(
            torch.Tensor(self.n_heads, self.input_dim, self.hidden_dim)
        )
        self.W_K = nn.Parameter(
            torch.Tensor(self.n_heads, self.input_dim, self.hidden_dim)
        )

        self.agg = MLP(input_dim=2 * self.n_heads + 4, output_dim=1, num_neurons=[32, 32])

        self.init_parameters()

    def init_parameters(self) -> None:
        for param in self.parameters():
            stdv = 1.0 / math.sqrt(param.size(-1))
            param.data.uniform_(-stdv, stdv)

    def forward(self, td: TensorDict, final_h: Tensor, final_p: Tensor) -> Tensor:
        """Compute the logits of the removing a node pair from the current solution

        Args:
            td: TensorDict with the current environment state
            final_h: final node embeddings
            final_p: final positional embeddings
        """

        selection_recent = torch.cat(
            (td["action_record"][:, -3:], td["action_record"].mean(1, True)), 1
        )
        solution = td["rec_current"]

        pre = solution.argsort()  # pre=[1,2,0]
        post = solution.gather(
            1, solution
        )  # post=[1,2,0] # the second neighbour works better
        batch_size, graph_size_plus1, input_dim = final_h.size()

        hflat = final_h.contiguous().view(-1, input_dim)  #################   reshape

        shp = (self.n_heads, batch_size, graph_size_plus1, self.hidden_dim)

        # Calculate queries, (n_heads, batch_size, graph_size+1, key_size)
        hidden_Q = torch.matmul(hflat, self.W_Q).view(shp)
        hidden_K = torch.matmul(hflat, self.W_K).view(shp)

        Q_pre = hidden_Q.gather(
            2, pre.view(1, batch_size, graph_size_plus1, 1).expand_as(hidden_Q)
        )
        K_post = hidden_K.gather(
            2, post.view(1, batch_size, graph_size_plus1, 1).expand_as(hidden_Q)
        )

        compatibility = (
            (Q_pre * hidden_K).sum(-1)
            + (hidden_Q * K_post).sum(-1)
            - (Q_pre * K_post).sum(-1)
        )[
            :, :, 1:
        ]  # (n_heads, batch_size, graph_size) (12)

        compatibility_pairing = torch.cat(
            (
                compatibility[:, :, : graph_size_plus1 // 2],
                compatibility[:, :, graph_size_plus1 // 2 :],
            ),
            0,
        )  # (n_heads*2, batch_size, graph_size/2)

        compatibility_pairing = self.agg(
            torch.cat(
                (
                    compatibility_pairing.permute(1, 2, 0),
                    selection_recent.permute(0, 2, 1),
                ),
                -1,
            )
        ).squeeze()  # (batch_size, graph_size/2)

        return compatibility_pairing


class NodePairReinsertionDecoder(ImprovementDecoder):
    """
    N2S Node-Pair Reinsertion decoder based on Ma et al. (2022)
    Given the environment state, the node embeddings (positional embeddings are discarded), and the removed node from the NodePairRemovalDecoder,
    compute the logits for finding places to re-insert the removed pair of pickup and delivery nodes to form a new solution


    Args:
        embed_dim: Embedding dimension
        num_heads: Number of attention heads
    """

    def __init__(
        self,
        embed_dim: int = 128,
        num_heads: int = 4,
    ):
        super().__init__()
        self.input_dim = embed_dim
        self.n_heads = num_heads
        self.hidden_dim = embed_dim

        assert embed_dim % num_heads == 0

        self.compater_insert1 = MultiHeadCompat(
            num_heads, embed_dim, embed_dim, embed_dim, embed_dim
        )

        self.compater_insert2 = MultiHeadCompat(
            num_heads, embed_dim, embed_dim, embed_dim, embed_dim
        )

        self.agg = MLP(input_dim=4 * self.n_heads, output_dim=1, num_neurons=[32, 32])

    def forward(self, td: TensorDict, final_h: Tensor, final_p: Tensor) -> torch.Tensor:
        action_removal = td["action"]
        solution = td["rec_current"]

        pos_pickup = (1 + action_removal).view(-1)
        pos_delivery = pos_pickup + solution.size(-1) // 2

        batch_size, graph_size_plus1, input_dim = final_h.size()
        shp = (batch_size, graph_size_plus1, graph_size_plus1, self.n_heads)
        shp_p = (batch_size, -1, 1, self.n_heads)
        shp_d = (batch_size, 1, -1, self.n_heads)

        arange = torch.arange(batch_size, device=final_h.device)
        h_pickup = final_h[arange, pos_pickup].unsqueeze(1)  # (batch_size, 1, input_dim)
        h_delivery = final_h[arange, pos_delivery].unsqueeze(
            1
        )  # (batch_size, 1, input_dim)
        h_K_neibour = final_h.gather(
            1, solution.view(batch_size, graph_size_plus1, 1).expand_as(final_h)
        )  # (batch_size, graph_size+1, input_dim)

        compatibility_pickup_pre = (
            self.compater_insert1(
                h_pickup, final_h
            )  # (n_heads, batch_size, 1, graph_size+1)
            .permute(1, 2, 3, 0)  # (batch_size, 1, graph_size+1, n_heads)
            .view(shp_p)  # (batch_size, graph_size+1, 1, n_heads)
            .expand(shp)  # (batch_size, graph_size+1, graph_size+1, n_heads)
        )
        compatibility_pickup_post = (
            self.compater_insert2(h_pickup, h_K_neibour)
            .permute(1, 2, 3, 0)
            .view(shp_p)
            .expand(shp)
        )
        compatibility_delivery_pre = (
            self.compater_insert1(
                h_delivery, final_h
            )  # (n_heads, batch_size, 1, graph_size+1)
            .permute(1, 2, 3, 0)  # (batch_size, 1, graph_size+1, n_heads)
            .view(shp_d)  # (batch_size, 1, graph_size+1, n_heads)
            .expand(shp)  # (batch_size, graph_size+1, graph_size+1, n_heads)
        )
        compatibility_delivery_post = (
            self.compater_insert2(h_delivery, h_K_neibour)
            .permute(1, 2, 3, 0)
            .view(shp_d)
            .expand(shp)
        )

        compatibility = self.agg(
            torch.cat(
                (
                    compatibility_pickup_pre,
                    compatibility_pickup_post,
                    compatibility_delivery_pre,
                    compatibility_delivery_post,
                ),
                -1,
            )
        ).squeeze()

        return compatibility  # (batch_size, graph_size+1, graph_size+1)


class CriticDecoder(nn.Module):
    def __init__(self, input_dim: int, dropout_rate=0.01) -> None:
        super().__init__()
        self.input_dim = input_dim

        self.project_graph = nn.Linear(self.input_dim, self.input_dim // 2)
        self.project_node = nn.Linear(self.input_dim, self.input_dim // 2)

        self.MLP = MLP(
            input_dim=input_dim + 1,
            output_dim=1,
            num_neurons=[input_dim, input_dim // 2],
            dropout_probs=[dropout_rate, 0.0],
        )

    def forward(self, x: torch.Tensor, best_cost: torch.Tensor) -> torch.Tensor:
        # h_wave: (batch_size, graph_size+1, input_size)
        mean_pooling = x.mean(1)  # mean Pooling (batch_size, input_size)
        graph_feature: torch.Tensor = self.project_graph(mean_pooling)[
            :, None, :
        ]  # (batch_size, 1, input_dim/2)
        node_feature: torch.Tensor = self.project_node(
            x
        )  # (batch_size, graph_size+1, input_dim/2)

        # pass through value_head, get estimated value
        fusion = node_feature + graph_feature.expand_as(
            node_feature
        )  # (batch_size, graph_size+1, input_dim/2)

        fusion_feature = torch.cat(
            (
                fusion.mean(1),
                fusion.max(1)[0],  # max_pooling
                best_cost.to(x.device),
            ),
            -1,
        )  # (batch_size, input_dim + 1)

        value = self.MLP(fusion_feature)

        return value



import math

import torch
import torch.nn as nn
import torch.nn.functional as F

from einops import einsum
from torch import Tensor

from rl4co.models.nn.env_embeddings import env_init_embedding
from rl4co.models.nn.ops import TransformerFFN


class HetGNNLayer(nn.Module):
    def __init__(
        self,
        embed_dim: int,
    ) -> None:
        super().__init__()

        self.self_attn = nn.Parameter(torch.rand(size=(embed_dim, 1), dtype=torch.float))
        self.cross_attn = nn.Parameter(torch.rand(size=(embed_dim, 1), dtype=torch.float))
        self.edge_attn = nn.Parameter(torch.rand(size=(embed_dim, 1), dtype=torch.float))
        self.activation = nn.ReLU()
        self.scale = 1 / math.sqrt(embed_dim)

    def forward(
        self, self_emb: Tensor, other_emb: Tensor, edge_emb: Tensor, edges: Tensor
    ):
        bs, n_rows, _ = self_emb.shape

        # concat operation embeddings and o-m edge features (proc times)
        # Calculate attention coefficients
        er = einsum(self_emb, self.self_attn, "b m e, e one -> b m") * self.scale
        ec = einsum(other_emb, self.cross_attn, "b o e, e one -> b o") * self.scale
        ee = einsum(edge_emb, self.edge_attn, "b m o e, e one -> b m o") * self.scale

        # element wise multiplication similar to broadcast column logits over rows with masking
        ec_expanded = einsum(edges, ec, "b m o, b o -> b m o")
        # element wise multiplication similar to broadcast row logits over cols with masking
        er_expanded = einsum(edges, er, "b m o, b m -> b m o")

        # adding the projections of different node types and edges together (equivalent to first concat and then project)
        # (bs, n_rows, n_cols)
        cross_logits = self.activation(ec_expanded + ee + er_expanded)

        # (bs, n_rows, 1)
        self_logits = self.activation(er + er).unsqueeze(-1)

        # (bs, n_ma, n_ops + 1)
        mask = torch.cat(
            (
                edges == 1,
                torch.full(
                    size=(bs, n_rows, 1),
                    dtype=torch.bool,
                    fill_value=True,
                    device=edges.device,
                ),
            ),
            dim=-1,
        )

        # (bs, n_ma, n_ops + 1)
        all_logits = torch.cat((cross_logits, self_logits), dim=-1)
        all_logits[~mask] = -torch.inf
        attn_scores = F.softmax(all_logits, dim=-1)
        # (bs, n_ma, n_ops)
        cross_attn_scores = attn_scores[..., :-1]
        # (bs, n_ma, 1)
        self_attn_scores = attn_scores[..., -1].unsqueeze(-1)

        # augment column embeddings with edge features, (bs, r, c, e)
        other_emb_aug = edge_emb + other_emb.unsqueeze(-3)
        cross_emb = einsum(cross_attn_scores, other_emb_aug, "b m o, b m o e -> b m e")
        self_emb = self_emb * self_attn_scores
        # (bs, n_ma, emb_dim)
        hidden = cross_emb + self_emb
        return hidden


class HetGNNBlock(nn.Module):
    def __init__(self, embed_dim, normalization: str = "batch") -> None:
        super().__init__()
        self.hgnn1 = HetGNNLayer(embed_dim)
        self.hgnn2 = HetGNNLayer(embed_dim)
        self.ffn1 = TransformerFFN(embed_dim, embed_dim * 2, normalization=normalization)
        self.ffn2 = TransformerFFN(embed_dim, embed_dim * 2, normalization=normalization)

    def forward(self, x1, x2, edge_emb, edges):
        h1 = self.hgnn1(x1, x2, edge_emb, edges)
        h1 = self.ffn1(h1, x1)

        h2 = self.hgnn2(x2, x1, edge_emb.transpose(1, 2), edges.transpose(1, 2))
        h2 = self.ffn2(h2, x2)

        return h1, h2


class HetGNNEncoder(nn.Module):
    def __init__(
        self,
        embed_dim: int,
        num_layers: int = 2,
        normalization: str = "batch",
        init_embedding=None,
        env_name: str = "fjsp",
        **init_embedding_kwargs,
    ) -> None:
        super().__init__()

        if init_embedding is None:
            init_embedding_kwargs["embed_dim"] = embed_dim
            init_embedding = env_init_embedding(env_name, init_embedding_kwargs)

        self.init_embedding = init_embedding

        self.num_layers = num_layers
        self.layers = nn.ModuleList(
            [HetGNNBlock(embed_dim, normalization) for _ in range(num_layers)]
        )

    def forward(self, td):
        row_emb, col_emb, edge_emb, edges = self.init_embedding(td)
        # perform sanity check to validate correct order of row and col embeddings
        n_rows, n_cols = edges.shape[1:]
        assert row_emb.size(1) == n_rows, "incorrect number of row embeddings"
        assert col_emb.size(1) == n_cols, "incorrect number of column embeddings"

        for layer in self.layers:
            row_emb, col_emb = layer(row_emb, col_emb, edge_emb, edges)

        return (row_emb, col_emb), None



from typing import Callable, Union

import torch
import torch.nn as nn

from torch import Tensor

try:
    from torch_geometric.data import Batch, Data
except ImportError:
    Batch = Data = None

from rl4co.utils.ops import get_distance_matrix, get_full_graph_edge_index, sparsify_graph
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


def env_edge_embedding(env_name: str, config: dict) -> nn.Module:
    """Retrieve the edge embedding module specific to the environment. Edge embeddings are crucial for
    transforming the raw edge features into a format suitable for the neural network, especially in
    graph neural networks where edge features can significantly impact the model's performance.

    Args:
        env: Environment or its name.
        config: A dictionary of configuration options for the environment.
    """
    embedding_registry = {
        "tsp": TSPEdgeEmbedding,
        "atsp": ATSPEdgeEmbedding,
        "cvrp": CVRPEdgeEmbedding,
        "cvrpmvc": CVRPEdgeEmbedding,
        "sdvrp": TSPEdgeEmbedding,
        "pctsp": CVRPEdgeEmbedding,
        "spctsp": TSPEdgeEmbedding,
        "op": CVRPEdgeEmbedding,
        "dpp": TSPEdgeEmbedding,
        "mdpp": TSPEdgeEmbedding,
        "pdp": TSPEdgeEmbedding,
        "mtsp": TSPEdgeEmbedding,
        "smtwtp": NoEdgeEmbedding,
        "shpp": TSPEdgeEmbedding,
    }

    if env_name not in embedding_registry:
        raise ValueError(
            f"Unknown environment name '{env_name}'. Available init embeddings: {embedding_registry.keys()}"
        )

    return embedding_registry[env_name](**config)


class TSPEdgeEmbedding(nn.Module):
    """Edge embedding module for the Traveling Salesman Problem (TSP) and related problems.
    This module converts the cost matrix or the distances between nodes into embeddings that can be
    used by the neural network. It supports sparsification to focus on a subset of relevant edges,
    which is particularly useful for large graphs.
    """

    node_dim = 1

    def __init__(
        self,
        embed_dim,
        linear_bias=True,
        sparsify=True,
        k_sparse: Union[int, Callable[[int], int], None] = None,
    ):
        assert Batch is not None, (
            "torch_geometric not found. Please install torch_geometric using instructions from "
            "https://pytorch-geometric.readthedocs.io/en/latest/install/installation.html."
        )

        super(TSPEdgeEmbedding, self).__init__()

        if k_sparse is None:
            self._get_k_sparse = lambda n: max(n // 5, 10)
        elif isinstance(k_sparse, int):
            self._get_k_sparse = lambda n: k_sparse
        elif callable(k_sparse):
            self._get_k_sparse = k_sparse
        else:
            raise ValueError("k_sparse must be an int or a callable")

        self.sparsify = sparsify
        self.edge_embed = nn.Linear(self.node_dim, embed_dim, linear_bias)

    def forward(self, td, init_embeddings: Tensor):
        cost_matrix = get_distance_matrix(td["locs"])
        batch = self._cost_matrix_to_graph(cost_matrix, init_embeddings)
        return batch

    def _cost_matrix_to_graph(self, batch_cost_matrix: Tensor, init_embeddings: Tensor):
        """Convert batched cost_matrix to batched PyG graph, and calculate edge embeddings.

        Args:
            batch_cost_matrix: Tensor of shape [batch_size, n, n]
            init_embedding: init embeddings
        """
        k_sparse = self._get_k_sparse(batch_cost_matrix.shape[-1])
        graph_data = []
        for index, cost_matrix in enumerate(batch_cost_matrix):
            if self.sparsify:
                edge_index, edge_attr = sparsify_graph(
                    cost_matrix, k_sparse, self_loop=False
                )
            else:
                edge_index = get_full_graph_edge_index(
                    cost_matrix.shape[0], self_loop=False
                ).to(cost_matrix.device)
                edge_attr = cost_matrix[edge_index[0], edge_index[1]].unsqueeze(-1)

            graph = Data(
                x=init_embeddings[index], edge_index=edge_index, edge_attr=edge_attr
            )
            graph_data.append(graph)

        batch = Batch.from_data_list(graph_data)
        batch.edge_attr = self.edge_embed(batch.edge_attr)
        return batch


class CVRPEdgeEmbedding(TSPEdgeEmbedding):
    """Edge embedding module for the Capacitated Vehicle Routing Problem (CVRP).
    Unlike the TSP, all nodes in the CVRP should be connected to the depot,
    so each node will have k_sparse + 1 edges.
    """

    def _cost_matrix_to_graph(self, batch_cost_matrix: Tensor, init_embeddings: Tensor):
        """Convert batched cost_matrix to batched PyG graph, and calculate edge embeddings.

        Args:
            batch_cost_matrix: Tensor of shape [batch_size, n, n]
            init_embedding: init embeddings
        """
        k_sparse = self._get_k_sparse(batch_cost_matrix.shape[-1])
        graph_data = []
        for index, cost_matrix in enumerate(batch_cost_matrix):
            if self.sparsify:
                edge_index, edge_attr = sparsify_graph(
                    cost_matrix[1:, 1:], k_sparse, self_loop=False
                )
                edge_index = edge_index + 1  # because we removed the depot
                # Note here
                edge_index = torch.cat(
                    [
                        edge_index,
                        # All nodes should be connected to the depot
                        torch.stack(
                            [
                                torch.arange(1, cost_matrix.shape[0]),
                                torch.zeros(cost_matrix.shape[0] - 1, dtype=torch.long),
                            ]
                        ).to(edge_index.device),
                        # Depot should be connected to all nodes
                        torch.stack(
                            [
                                torch.zeros(cost_matrix.shape[0] - 1, dtype=torch.long),
                                torch.arange(1, cost_matrix.shape[0]),
                            ]
                        ).to(edge_index.device),
                    ],
                    dim=1,
                )
                edge_attr = torch.cat(
                    [edge_attr, cost_matrix[1:, [0]], cost_matrix[[0], 1:].t()], dim=0
                )

            else:
                edge_index = get_full_graph_edge_index(
                    cost_matrix.shape[0], self_loop=False
                ).to(cost_matrix.device)
                edge_attr = cost_matrix[edge_index[0], edge_index[1]].unsqueeze(-1)

            graph = Data(
                x=init_embeddings[index], edge_index=edge_index, edge_attr=edge_attr
            )
            graph_data.append(graph)

        batch = Batch.from_data_list(graph_data)  # type: ignore
        batch.edge_attr = self.edge_embed(batch.edge_attr)  # type: ignore
        return batch


class VRPPolarEdgeEmbedding(TSPEdgeEmbedding):
    """Edge embedding module for VRPs using polar coordinates, with the depot as the origin.

    This class extends TSPEdgeEmbedding to handle VRP problems by incorporating polar coordinate
    information for edge embeddings.
    """

    node_dim = 2

    def forward(self, td, init_embeddings: Tensor):
        with torch.no_grad():
            if "polar_locs" in td.keys():
                theta = td["polar_locs"][..., 1]
            else:
                shifted_locs = td["locs"] - td["locs"][..., 0:1, :]
                x, y = shifted_locs[..., 0], shifted_locs[..., 1]
                theta = torch.atan2(y, x)

            delta_theta_matrix = theta[..., :, None] - theta[..., None, :]
            edge_attr1 = 1 - torch.cos(delta_theta_matrix)
            edge_attr2 = get_distance_matrix(td["locs"])
            cost_matrix = torch.stack((edge_attr1, edge_attr2), dim=-1)
            del edge_attr1, edge_attr2, delta_theta_matrix

            batch = self._cost_matrix_to_graph(cost_matrix, init_embeddings)
            del cost_matrix

        batch.edge_attr = self.edge_embed(batch.edge_attr)  # type: ignore
        return batch

    @torch.no_grad()
    def _cost_matrix_to_graph(self, batch_cost_matrix: Tensor, init_embeddings: Tensor):
        """Convert batched cost_matrix to batched PyG graph, and calculate edge embeddings.

        Args:
            batch_cost_matrix: Tensor of shape [batch_size, n, n, m]
            init_embedding: init embeddings of shape [batch_size, n, m]
        """
        k_sparse = self._get_k_sparse(batch_cost_matrix.shape[-1])
        graph_data = []
        for index, cost_matrix in enumerate(batch_cost_matrix):
            edge_index, _ = sparsify_graph(cost_matrix[..., 0], k_sparse, self_loop=False)
            edge_index = edge_index.T[torch.all(edge_index != 0, dim=0)].T
            _, depot_edge_index = torch.topk(
                cost_matrix[0, :, 1], k=k_sparse, largest=False, sorted=False
            )
            depot_edge_index = depot_edge_index[depot_edge_index != 0]
            depot_edge_index = torch.stack(
                (torch.zeros_like(depot_edge_index), depot_edge_index), dim=0
            )
            edge_index = torch.concat((depot_edge_index, edge_index), dim=-1).detach()
            edge_attr = cost_matrix[edge_index[0], edge_index[1]].detach()

            graph = Data(
                x=init_embeddings[index],
                edge_index=edge_index,
                edge_attr=edge_attr,
            )  # type: ignore
            graph_data.append(graph)

        batch = Batch.from_data_list(graph_data)  # type: ignore
        return batch


class ATSPEdgeEmbedding(TSPEdgeEmbedding):
    """Edge embedding module for the Asymmetric Traveling Salesman Problem (ATSP).
    Inherits from TSPEdgeEmbedding and adapts the edge embedding process to handle
    asymmetric cost matrices, where the cost from node i to node j may not be the same as from j to i.
    """

    def forward(self, td, init_embeddings: Tensor):
        batch = self._cost_matrix_to_graph(td["cost_matrix"], init_embeddings)
        return batch


class NoEdgeEmbedding(nn.Module):
    """A module for environments that do not require edge embeddings, or where edge features
    are not used. This can be useful for simplifying models in problems where only node
    features are relevant.
    """

    def __init__(self, embed_dim, self_loop=False, **kwargs):
        assert Batch is not None, (
            "torch_geometric not found. Please install torch_geometric using instructions from "
            "https://pytorch-geometric.readthedocs.io/en/latest/install/installation.html."
        )

        super(NoEdgeEmbedding, self).__init__()
        self.embed_dim = embed_dim
        self.self_loop = self_loop

    def forward(self, td, init_embeddings: Tensor):
        data_list = []
        n = init_embeddings.shape[1]
        device = init_embeddings.device
        edge_index = get_full_graph_edge_index(n, self_loop=self.self_loop).to(device)
        m = edge_index.shape[1]

        for node_embed in init_embeddings:
            data = Data(
                x=node_embed,
                edge_index=edge_index,
                edge_attr=torch.zeros((m, self.embed_dim), device=device),
            )  # type: ignore
            data_list.append(data)

        batch = Batch.from_data_list(data_list)  # type: ignore
        return batch



from rl4co.models.common.constructive.autoregressive import (
    AutoregressiveDecoder,
    AutoregressiveEncoder,
    AutoregressivePolicy,
)
from rl4co.models.common.constructive.base import (
    ConstructiveDecoder,
    ConstructiveEncoder,
    ConstructivePolicy,
)
from rl4co.models.common.constructive.nonautoregressive import (
    NonAutoregressiveDecoder,
    NonAutoregressiveEncoder,
    NonAutoregressivePolicy,
)
from rl4co.models.common.transductive import TransductiveModel
from rl4co.models.rl import StepwisePPO
from rl4co.models.rl.a2c.a2c import A2C
from rl4co.models.rl.common.base import RL4COLitModule
from rl4co.models.rl.ppo.ppo import PPO
from rl4co.models.rl.reinforce.baselines import REINFORCEBaseline, get_reinforce_baseline
from rl4co.models.rl.reinforce.reinforce import REINFORCE
from rl4co.models.zoo.active_search import ActiveSearch
from rl4co.models.zoo.am import AttentionModel, AttentionModelPolicy
from rl4co.models.zoo.amppo import AMPPO
from rl4co.models.zoo.dact import DACT, DACTPolicy
from rl4co.models.zoo.deepaco import DeepACO, DeepACOPolicy
from rl4co.models.zoo.eas import EAS, EASEmb, EASLay
from rl4co.models.zoo.glop import GLOP, GLOPPolicy
from rl4co.models.zoo.gfacs import GFACS, GFACSPolicy
from rl4co.models.zoo.ham import (
    HeterogeneousAttentionModel,
    HeterogeneousAttentionModelPolicy,
)
from rl4co.models.zoo.l2d import (
    L2DAttnPolicy,
    L2DModel,
    L2DPolicy,
    L2DPolicy4PPO,
    L2DPPOModel,
)
from rl4co.models.zoo.matnet import MatNet, MatNetPolicy
from rl4co.models.zoo.mdam import MDAM, MDAMPolicy
from rl4co.models.zoo.mvmoe import MVMoE_AM, MVMoE_POMO
from rl4co.models.zoo.n2s import N2S, N2SPolicy
from rl4co.models.zoo.nargnn import NARGNNPolicy
from rl4co.models.zoo.neuopt import NeuOpt, NeuOptPolicy
from rl4co.models.zoo.polynet import PolyNet
from rl4co.models.zoo.pomo import POMO
from rl4co.models.zoo.ptrnet import PointerNetwork, PointerNetworkPolicy
from rl4co.models.zoo.symnco import SymNCO, SymNCOPolicy



import copy

from typing import Any

import torch
import torch.nn as nn
import torch.nn.functional as F

from torchrl.data.replay_buffers import (
    LazyMemmapStorage,
    ListStorage,
    SamplerWithoutReplacement,
    TensorDictReplayBuffer,
)

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.models.rl.common.base import RL4COLitModule
from rl4co.models.rl.common.utils import RewardScaler
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


def make_replay_buffer(buffer_size, batch_size, device="cpu"):
    if device == "cpu":
        storage = LazyMemmapStorage(buffer_size, device="cpu")
        prefetch = 3
    else:
        storage = ListStorage(buffer_size)
        prefetch = None
    return TensorDictReplayBuffer(
        storage=storage,
        batch_size=batch_size,
        sampler=SamplerWithoutReplacement(drop_last=True),
        pin_memory=False,
        prefetch=prefetch,
    )


class StepwisePPO(RL4COLitModule):
    def __init__(
        self,
        env: RL4COEnvBase,
        policy: nn.Module,
        clip_range: float = 0.2,  # epsilon of PPO
        update_timestep: int = 1,
        buffer_size: int = 100_000,
        ppo_epochs: int = 2,  # inner epoch, K
        batch_size: int = 256,
        mini_batch_size: int = 256,
        vf_lambda: float = 0.5,  # lambda of Value function fitting
        entropy_lambda: float = 0.01,  # lambda of entropy bonus
        max_grad_norm: float = 0.5,  # max gradient norm
        buffer_storage_device: str = "gpu",
        metrics: dict = {
            "train": ["loss", "surrogate_loss", "value_loss", "entropy"],
        },
        reward_scale: str | int = None,
        **kwargs,
    ):
        super().__init__(env, policy, metrics=metrics, batch_size=batch_size, **kwargs)

        self.policy_old = copy.deepcopy(self.policy)
        self.automatic_optimization = False  # PPO uses custom optimization routine
        self.rb = make_replay_buffer(buffer_size, mini_batch_size, buffer_storage_device)
        self.scaler = RewardScaler(reward_scale)

        self.ppo_cfg = {
            "clip_range": clip_range,
            "ppo_epochs": ppo_epochs,
            "update_timestep": update_timestep,
            "mini_batch_size": mini_batch_size,
            "vf_lambda": vf_lambda,
            "entropy_lambda": entropy_lambda,
            "max_grad_norm": max_grad_norm,
        }

    def update(self, device):
        outs = []
        # PPO inner epoch
        for _ in range(self.ppo_cfg["ppo_epochs"]):
            for sub_td in self.rb:
                sub_td = sub_td.to(device)
                previous_reward = sub_td["reward"].view(-1, 1)
                previous_logp = sub_td["logprobs"]

                logprobs, value_pred, entropy = self.policy.evaluate(sub_td)

                ratios = torch.exp(logprobs - previous_logp)

                advantages = torch.squeeze(previous_reward - value_pred.detach(), 1)
                surr1 = ratios * advantages
                surr2 = (
                    torch.clamp(
                        ratios,
                        1 - self.ppo_cfg["clip_range"],
                        1 + self.ppo_cfg["clip_range"],
                    )
                    * advantages
                )
                surrogate_loss = -torch.min(surr1, surr2).mean()

                # compute value function loss
                value_loss = F.mse_loss(value_pred, previous_reward)

                # compute total loss
                loss = (
                    surrogate_loss
                    + self.ppo_cfg["vf_lambda"] * value_loss
                    - self.ppo_cfg["entropy_lambda"] * entropy.mean()
                )

                # perform manual optimization following the Lightning routine
                # https://lightning.ai/docs/pytorch/stable/common/optimization.html

                opt = self.optimizers()
                opt.zero_grad()
                self.manual_backward(loss)
                if self.ppo_cfg["max_grad_norm"] is not None:
                    self.clip_gradients(
                        opt,
                        gradient_clip_val=self.ppo_cfg["max_grad_norm"],
                        gradient_clip_algorithm="norm",
                    )
                opt.step()

                out = {
                    "reward": previous_reward.mean(),
                    "loss": loss,
                    "surrogate_loss": surrogate_loss,
                    "value_loss": value_loss,
                    "entropy": entropy.mean(),
                }

                outs.append(out)
        # Copy new weights into old policy:
        self.policy_old.load_state_dict(self.policy.state_dict())
        outs = {k: torch.stack([dic[k] for dic in outs], dim=0) for k in outs[0]}
        return outs

    def shared_step(
        self, batch: Any, batch_idx: int, phase: str, dataloader_idx: int = None
    ):
        next_td = self.env.reset(batch)
        device = next_td.device
        if phase == "train":
            while not next_td["done"].all():
                with torch.no_grad():
                    td = self.policy_old.act(next_td, self.env, phase="train")
                # get next state
                next_td = self.env.step(td)["next"]
                # get reward of action
                reward = self.env.get_reward(next_td, None)
                reward = self.scaler(reward)
                # add reward to prior state
                td.set("reward", reward)
                # add tensordict with action, logprobs and reward information to buffer
                self.rb.extend(td)

            # if iter mod x = 0 then update the policy (x = 1 in paper)
            if batch_idx % self.ppo_cfg["update_timestep"] == 0:
                out = self.update(device)
                self.rb.empty()

        else:
            out = self.policy.generate(
                next_td, self.env, phase=phase, select_best=phase != "train"
            )

        metrics = self.log_metrics(out, phase, dataloader_idx=dataloader_idx)
        return {"loss": out.get("loss", None), **metrics}



from typing import Optional

import torch

from tensordict.tensordict import TensorDict
from torchrl.data import Bounded, Composite, Unbounded

from rl4co.data.utils import load_npz_to_tensordict
from rl4co.envs.common.base import RL4COEnvBase
from rl4co.utils.ops import gather_by_index, get_tour_length
from rl4co.utils.pylogger import get_pylogger

from .generator import CVRPGenerator

try:
    from .local_search import local_search
except:  # In case when we fail to build HGS
    local_search = None
from .render import render

log = get_pylogger(__name__)


class CVRPEnv(RL4COEnvBase):
    """Capacitated Vehicle Routing Problem (CVRP) environment.
    At each step, the agent chooses a customer to visit depending on the current location and the remaining capacity.
    When the agent visits a customer, the remaining capacity is updated. If the remaining capacity is not enough to
    visit any customer, the agent must go back to the depot. The reward is 0 unless the agent visits all the cities.
    In that case, the reward is (-)length of the path: maximizing the reward is equivalent to minimizing the path length.

    Observations:
        - location of the depot.
        - locations and demand of each customer.
        - current location of the vehicle.
        - the remaining customer of the vehicle,

    Constraints:
        - the tour starts and ends at the depot.
        - each customer must be visited exactly once.
        - the vehicle cannot visit customers exceed the remaining capacity.
        - the vehicle can return to the depot to refill the capacity.

    Finish Condition:
        - the vehicle has visited all customers and returned to the depot.

    Reward:
        - (minus) the negative length of the path.

    Args:
        generator: CVRPGenerator instance as the data generator
        generator_params: parameters for the generator
    """

    name = "cvrp"

    def __init__(
        self,
        generator: CVRPGenerator = None,
        generator_params: dict = {},
        **kwargs,
    ):
        super().__init__(**kwargs)
        if generator is None:
            generator = CVRPGenerator(**generator_params)
        self.generator = generator
        self._make_spec(self.generator)

    def _step(self, td: TensorDict) -> TensorDict:
        current_node = td["action"][:, None]  # Add dimension for step
        n_loc = td["demand"].size(-1)  # Excludes depot

        # Not selected_demand is demand of first node (by clamp) so incorrect for nodes that visit depot!
        selected_demand = gather_by_index(
            td["demand"], torch.clamp(current_node - 1, 0, n_loc - 1), squeeze=False
        )

        # Increase capacity if depot is not visited, otherwise set to 0
        used_capacity = (td["used_capacity"] + selected_demand) * (
            current_node != 0
        ).float()

        # Note: here we do not subtract one as we have to scatter so the first column allows scattering depot
        # Add one dimension since we write a single value
        visited = td["visited"].scatter(-1, current_node, 1)

        # SECTION: get done
        done = visited.sum(-1) == visited.size(-1)
        reward = torch.zeros_like(done)

        td.update(
            {
                "current_node": current_node,
                "used_capacity": used_capacity,
                "visited": visited,
                "reward": reward,
                "done": done,
            }
        )
        td.set("action_mask", self.get_action_mask(td))
        return td

    def _reset(
        self,
        td: Optional[TensorDict] = None,
        batch_size: Optional[list] = None,
    ) -> TensorDict:
        device = td.device

        # Create reset TensorDict
        td_reset = TensorDict(
            {
                "locs": torch.cat((td["depot"][:, None, :], td["locs"]), -2),
                "demand": td["demand"],
                "current_node": torch.zeros(
                    *batch_size, 1, dtype=torch.long, device=device
                ),
                "used_capacity": torch.zeros((*batch_size, 1), device=device),
                "vehicle_capacity": torch.full(
                    (*batch_size, 1), self.generator.vehicle_capacity, device=device
                ),
                "visited": torch.zeros(
                    (*batch_size, td["locs"].shape[-2] + 1),
                    dtype=torch.uint8,
                    device=device,
                ),
            },
            batch_size=batch_size,
        )
        td_reset.set("action_mask", self.get_action_mask(td_reset))
        return td_reset

    @staticmethod
    def get_action_mask(td: TensorDict) -> torch.Tensor:
        # For demand steps_dim is inserted by indexing with id, for used_capacity insert node dim for broadcasting
        exceeds_cap = td["demand"] + td["used_capacity"] > td["vehicle_capacity"] + 1e-5

        # Nodes that cannot be visited are already visited or too much demand to be served now
        mask_loc = td["visited"][..., 1:].to(exceeds_cap.dtype) | exceeds_cap

        # Cannot visit the depot if just visited and still unserved nodes
        mask_depot = (td["current_node"] == 0) & ((mask_loc == 0).int().sum(-1) > 0)[
            :, None
        ]
        return ~torch.cat((mask_depot, mask_loc), -1)

    def _get_reward(self, td: TensorDict, actions: TensorDict) -> TensorDict:
        # Gather locations in order of tour (add depot since we start and end there)
        locs_ordered = torch.cat(
            [
                td["locs"][..., 0:1, :],  # depot
                gather_by_index(td["locs"], actions),  # order locations
            ],
            dim=1,
        )
        return -get_tour_length(locs_ordered)

    @staticmethod
    def check_solution_validity(td: TensorDict, actions: torch.Tensor):
        """Check that solution is valid: nodes are not visited twice except depot and capacity is not exceeded"""
        # Check if tour is valid, i.e. contain 0 to n-1
        batch_size, graph_size = td["demand"].size()
        sorted_pi = actions.data.sort(1)[0]

        # Sorting it should give all zeros at front and then 1...n
        assert (
            torch.arange(1, graph_size + 1, out=sorted_pi.data.new())
            .view(1, -1)
            .expand(batch_size, graph_size)
            == sorted_pi[:, -graph_size:]
        ).all() and (sorted_pi[:, :-graph_size] == 0).all(), "Invalid tour"

        # Visiting depot resets capacity so we add demand = -capacity (we make sure it does not become negative)
        demand_with_depot = torch.cat((-td["vehicle_capacity"], td["demand"]), 1)
        d = demand_with_depot.gather(1, actions)

        used_cap = torch.zeros_like(td["demand"][:, 0])
        for i in range(actions.size(1)):
            used_cap += d[
                :, i
            ]  # This will reset/make capacity negative if i == 0, e.g. depot visited
            # Cannot use less than 0
            used_cap[used_cap < 0] = 0
            assert (
                used_cap <= td["vehicle_capacity"][:, 0] + 1e-5
            ).all(), "Used more than capacity"

    @staticmethod
    def load_data(fpath, batch_size=[]):
        """Dataset loading from file
        Normalize demand by capacity to be in [0, 1]
        """
        td_load = load_npz_to_tensordict(fpath)
        td_load.set("demand", td_load["demand"] / td_load["capacity"][:, None])
        return td_load

    def _make_spec(self, generator: CVRPGenerator):
        self.observation_spec = Composite(
            locs=Bounded(
                low=generator.min_loc,
                high=generator.max_loc,
                shape=(generator.num_loc + 1, 2),
                dtype=torch.float32,
            ),
            current_node=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            demand=Bounded(
                low=-generator.capacity,
                high=generator.max_demand,
                shape=(generator.num_loc + 1, 1),
                dtype=torch.float32,
            ),
            action_mask=Unbounded(
                shape=(generator.num_loc + 1, 1),
                dtype=torch.bool,
            ),
            shape=(),
        )
        self.action_spec = Bounded(
            shape=(1,),
            dtype=torch.int64,
            low=0,
            high=generator.num_loc + 1,
        )
        self.reward_spec = Unbounded(shape=(1,))
        self.done_spec = Unbounded(shape=(1,), dtype=torch.bool)

    def replace_selected_actions(
        self,
        cur_actions: torch.Tensor,
        new_actions: torch.Tensor,
        selection_mask: torch.Tensor,
    ) -> torch.Tensor:
        """
        Replace selected current actions with updated actions based on `selection_mask`.

        Args:
            cur_actions [batch_size, num_loc]
            new_actions [batch_size, num_loc]
            selection_mask [batch_size,]
        """
        diff_length = cur_actions.size(-1) - new_actions.size(-1)
        if diff_length > 0:
            new_actions = torch.nn.functional.pad(
                new_actions, (0, diff_length, 0, 0), mode="constant", value=0
            )
        elif diff_length < 0:
            cur_actions = torch.nn.functional.pad(
                cur_actions, (0, -diff_length, 0, 0), mode="constant", value=0
            )
        cur_actions[selection_mask] = new_actions[selection_mask]
        return cur_actions

    @staticmethod
    def local_search(td: TensorDict, actions: torch.Tensor, **kwargs) -> torch.Tensor:
        assert (
            local_search is not None
        ), "Cannot import local_search module. Check `rl4co/envs/routing/cvrp/README.md` for instructions to build HGS."
        return local_search(td, actions, **kwargs)

    @staticmethod
    def render(td: TensorDict, actions: torch.Tensor = None, ax=None):
        return render(td, actions, ax)



from dataclasses import dataclass
from typing import Tuple

import torch.nn as nn

from torch import Tensor

from rl4co.envs import RL4COEnvBase
from rl4co.models.nn.attention import PolyNetAttention
from rl4co.models.nn.env_embeddings import env_context_embedding, env_dynamic_embedding
from rl4co.models.nn.env_embeddings.dynamic import StaticEmbedding
from rl4co.models.zoo.am.decoder import AttentionModelDecoder
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


@dataclass
class PrecomputedCache:
    node_embeddings: Tensor
    graph_context: Tensor | float
    glimpse_key: Tensor
    glimpse_val: Tensor
    logit_key: Tensor


class PolyNetDecoder(AttentionModelDecoder):
    """
    PolyNet decoder for constructing diverse solutions for combinatorial optimization problems.
    Given the environment state and the embeddings, compute the logits and sample actions autoregressively until
    all the environments in the batch have reached a terminal state.
    We additionally include support for multi-starts as it is more efficient to do so in the decoder as we can
    natively perform the attention computation.

    Args:
        k: Number of strategies to learn ("K" in the PolyNet paper)
        encoder_type: Type of encoder that should be used. "AM" or "MatNet" are supported
        embed_dim: Embedding dimension
        poly_layer_dim: Dimension of the PolyNet layers
        num_heads: Number of attention heads
        env_name: Name of the environment used to initialize embeddings
        context_embedding: Context embedding module
        dynamic_embedding: Dynamic embedding module
        mask_inner: Whether to mask the inner loop
        out_bias_pointer_attn: Whether to use a bias in the pointer attention
        linear_bias: Whether to use a bias in the linear layer
        use_graph_context: Whether to use the graph context
        check_nan: Whether to check for nan values during decoding
        sdpa_fn: scaled_dot_product_attention function
    """

    def __init__(
        self,
        k: int,
        encoder_type: str,
        embed_dim: int = 128,
        poly_layer_dim: int = 256,
        num_heads: int = 8,
        env_name: str | RL4COEnvBase = "tsp",
        context_embedding: nn.Module = None,
        dynamic_embedding: nn.Module = None,
        mask_inner: bool = True,
        out_bias_pointer_attn: bool = False,
        linear_bias: bool = False,
        use_graph_context: bool = True,
        check_nan: bool = True,
        sdpa_fn: callable = None,
        **unused_kwargs,
    ):
        super().__init__()

        if isinstance(env_name, RL4COEnvBase):
            env_name = env_name.name
        self.env_name = env_name
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.encoder_type = encoder_type

        assert embed_dim % num_heads == 0

        self.context_embedding = (
            env_context_embedding(self.env_name, {"embed_dim": embed_dim})
            if context_embedding is None
            else context_embedding
        )
        self.dynamic_embedding = (
            env_dynamic_embedding(self.env_name, {"embed_dim": embed_dim})
            if dynamic_embedding is None
            else dynamic_embedding
        )
        self.is_dynamic_embedding = (
            False if isinstance(self.dynamic_embedding, StaticEmbedding) else True
        )

        # MHA with Pointer mechanism (https://arxiv.org/abs/1506.03134)
        self.pointer = PolyNetAttention(
            k,
            embed_dim,
            poly_layer_dim,
            num_heads,
            mask_inner=mask_inner,
            out_bias=out_bias_pointer_attn,
            check_nan=check_nan,
            sdpa_fn=sdpa_fn,
        )

        # For each node we compute (glimpse key, glimpse value, logit key) so 3 * embed_dim
        self.project_node_embeddings = nn.Linear(
            embed_dim, 3 * embed_dim, bias=linear_bias
        )
        self.project_fixed_context = nn.Linear(embed_dim, embed_dim, bias=linear_bias)
        self.use_graph_context = use_graph_context

    def _precompute_cache_matnet(
        self, embeddings: Tuple[Tensor, Tensor], *args, **kwargs
    ):
        col_emb, row_emb = embeddings
        (
            glimpse_key_fixed,
            glimpse_val_fixed,
            logit_key,
        ) = self.project_node_embeddings(
            col_emb
        ).chunk(3, dim=-1)

        # Optionally disable the graph context from the initial embedding as done in POMO
        if self.use_graph_context:
            graph_context = self.project_fixed_context(col_emb.mean(1))
        else:
            graph_context = 0

        # Organize in a dataclass for easy access
        return PrecomputedCache(
            node_embeddings=row_emb,
            graph_context=graph_context,
            glimpse_key=glimpse_key_fixed,
            glimpse_val=glimpse_val_fixed,
            logit_key=logit_key,
        )

    def _precompute_cache(self, embeddings: Tuple[Tensor, Tensor], *args, **kwargs):
        if self.encoder_type == "AM":
            return super()._precompute_cache(embeddings, *args, **kwargs)
        elif self.encoder_type == "MatNet":
            return self._precompute_cache_matnet(embeddings, *args, **kwargs)



import torch
import torch.nn as nn
from torch.distributions.normal import Normal

from rl4co.models.nn.mlp import MLP

"""
    Pytorch Implementation based on <https://arxiv.org/abs/1701.06538>
    Author: David Rau
    Link: <https://github.com/davidmrau/mixture-of-experts/blob/master/moe.py>
"""


class SparseDispatcher(object):
    """
    Helper for implementing a mixture of experts.
    The purpose of this class is to create input minibatches for the experts
    and to combine the results of the experts to form a unified output tensor.

    There are two functions:
        dispatch - take an input Tensor and create input Tensors for each expert.
        combine - take output Tensors from each expert and form a combined output
          Tensor.  Outputs from different experts for the same batch element are
          summed together, weighted by the provided "gates".

    The class is initialized with a "gates" Tensor, which specifies which
    batch elements go to which experts, and the weights to use when combining
    the outputs.  Batch element b is sent to expert e iff gates[b, e] != 0.
    The inputs and outputs are all two-dimensional [batch, depth].
    Caller is responsible for collapsing additional dimensions prior to
    calling this class and reshaping the output to the original shape.
    See common_layers.reshape_like().

    Example use:
        gates: a float32 `Tensor` with shape `[batch_size, num_experts]`
        inputs: a float32 `Tensor` with shape `[batch_size, input_size]`
        experts: a list of length `num_experts` containing sub-networks.
    dispatcher = SparseDispatcher(num_experts, gates)
    expert_inputs = dispatcher.dispatch(inputs)
    expert_outputs = [experts[i](expert_inputs[i]) for i in range(num_experts)]
    outputs = dispatcher.combine(expert_outputs)
    The preceding code sets the output for a particular example b to:
    output[b] = Sum_i(gates[b, i] * experts[i](inputs[b]))
    This class takes advantage of sparsity in the gate matrix by including in the
    `Tensor`s for expert i only the batch elements for which `gates[b, i] > 0`.
    """

    def __init__(self, num_experts, gates):
        """Create a SparseDispatcher."""

        self._gates = gates
        self._num_experts = num_experts
        # sort experts
        sorted_experts, index_sorted_experts = torch.nonzero(gates).sort(0)
        # drop indices
        _, self._expert_index = sorted_experts.split(1, dim=1)
        # get according batch index for each expert
        self._batch_index = torch.nonzero(gates)[index_sorted_experts[:, 1], 0]
        # calculate num samples that each expert gets
        self._part_sizes = (gates > 0).sum(0).tolist()
        # expand gates to match with self._batch_index
        gates_exp = gates[self._batch_index.flatten()]
        self._nonzero_gates = torch.gather(gates_exp, 1, self._expert_index)

    def dispatch(self, inp):
        """Create one input Tensor for each expert.
        The `Tensor` for a expert `i` contains the slices of `inp` corresponding
        to the batch elements `b` where `gates[b, i] > 0`.
        Args:
            inp: a `Tensor` of shape "[batch_size, <extra_input_dims>]`
        Returns:
            a list of `num_experts` `Tensor`s with shapes
                [expert_batch_size_i, <extra_input_dims>]`.
        """

        # assigns samples to experts whose gate is nonzero
        # expand according to batch index so we can just split by _part_sizes
        inp_exp = inp[self._batch_index].squeeze(1)
        return torch.split(inp_exp, self._part_sizes, dim=0)

    def combine(self, expert_out, multiply_by_gates=True):
        """Sum together the expert output, weighted by the gates.
        The slice corresponding to a particular batch element `b` is computed
        as the sum over all experts `i` of the expert output, weighted by the
        corresponding gate values.  If `multiply_by_gates` is set to False, the
        gate values are ignored.
        Args:
            expert_out: a list of `num_experts` `Tensor`s, each with shape
                [expert_batch_size_i, <extra_output_dims>]`.
            multiply_by_gates: a boolean
        Returns:
            a `Tensor` with shape `[batch_size, <extra_output_dims>]`.
        """
        # apply exp to expert outputs, so we are not longer in log space
        stitched = torch.cat(expert_out, 0)

        if multiply_by_gates:
            stitched = stitched.mul(self._nonzero_gates)
        zeros = torch.zeros(self._gates.size(0), expert_out[-1].size(-1), requires_grad=True, device=stitched.device)
        # combine samples that have been processed by the same k experts
        combined = zeros.index_add(0, self._batch_index, stitched.float())
        return combined

    def expert_to_gates(self):
        """Gate values corresponding to the examples in the per-expert `Tensor`s.
        Returns:
            a list of `num_experts` one-dimensional `Tensor`s with type `tf.float32`
                and shapes `[expert_batch_size_i]`
        """
        # split nonzero gates for each expert
        return torch.split(self._nonzero_gates, self._part_sizes, dim=0)


class MoE(nn.Module):
    """Call a Sparsely gated mixture of experts layer with 1-layer Feed-Forward networks as experts.
    Args:
        input_size: integer - size of the input
        output_size: integer - size of the input
        num_experts: an integer - number of experts
        num_neurons: a list - hidden dimension of the experts
        noisy_gating: a boolean
        k: an integer - how many experts to use for each batch element
    """

    def __init__(self, input_size, output_size, num_neurons=[], hidden_act="ReLU", out_bias=True, num_experts=4, k=2, noisy_gating=True, **kwargs):
        super(MoE, self).__init__()
        self.noisy_gating = noisy_gating
        self.num_experts = num_experts
        self.output_size = output_size
        self.input_size = input_size
        self.k = k

        # instantiate experts
        if num_neurons != []:
            self.experts = nn.ModuleList([MLP(input_dim=input_size, output_dim=output_size, num_neurons=num_neurons,
                                              hidden_act=hidden_act) for _ in range(self.num_experts)])
        else:
            self.experts = nn.ModuleList([nn.Linear(self.input_size, self.output_size, bias=out_bias) for _ in range(self.num_experts)])
        self.w_gate = nn.Parameter(torch.zeros(input_size, num_experts), requires_grad=True)
        self.w_noise = nn.Parameter(torch.zeros(input_size, num_experts), requires_grad=True)

        self.softplus = nn.Softplus()
        self.softmax = nn.Softmax(-1)
        self.register_buffer("mean", torch.tensor([0.0]))
        self.register_buffer("std", torch.tensor([1.0]))
        assert(self.k <= self.num_experts)

    def cv_squared(self, x):
        """The squared coefficient of variation of a sample.
        Useful as a loss to encourage a positive distribution to be more uniform.
        Epsilons added for numerical stability.
        Returns 0 for an empty Tensor.
        Args:
            x: a `Tensor`.
        Returns:
            a `Scalar`.
        """
        eps = 1e-10
        # if only num_experts = 1

        if x.shape[0] == 1:
            return torch.tensor([0], device=x.device, dtype=x.dtype)
        return x.float().var() / (x.float().mean()**2 + eps)

    def _gates_to_load(self, gates):
        """Compute the true load per expert, given the gates.
        The load is the number of examples for which the corresponding gate is >0.
        Args:
            gates: a `Tensor` of shape [batch_size, n]
        Returns:
            a float32 `Tensor` of shape [n]
        """
        return (gates > 0).sum(0)

    def _prob_in_top_k(self, clean_values, noisy_values, noise_stddev, noisy_top_values):
        """Helper function to NoisyTopKGating.
        Computes the probability that value is in top k, given different random noise.
        This gives us a way of backpropagating from a loss that balances the number
        of times each expert is in the top k experts per example.
        In the case of no noise, pass in None for noise_stddev, and the result will
        not be differentiable.
        Args:
            clean_values: a `Tensor` of shape [batch, n].
            noisy_values: a `Tensor` of shape [batch, n].  Equal to clean values plus
              normally distributed noise with standard deviation noise_stddev.
            noise_stddev: a `Tensor` of shape [batch, n], or None
            noisy_top_values: a `Tensor` of shape [batch, m].
               "values" Output of tf.top_k(noisy_top_values, m).  m >= k+1
        Returns:
            a `Tensor` of shape [batch, n].
        """
        batch = clean_values.size(0)
        m = noisy_top_values.size(1)
        top_values_flat = noisy_top_values.flatten()

        threshold_positions_if_in = torch.arange(batch, device=clean_values.device) * m + self.k
        threshold_if_in = torch.unsqueeze(torch.gather(top_values_flat, 0, threshold_positions_if_in), 1)
        is_in = torch.gt(noisy_values, threshold_if_in)
        threshold_positions_if_out = threshold_positions_if_in - 1
        threshold_if_out = torch.unsqueeze(torch.gather(top_values_flat, 0, threshold_positions_if_out), 1)
        # is each value currently in the top k.
        normal = Normal(self.mean, self.std)
        prob_if_in = normal.cdf((clean_values - threshold_if_in)/noise_stddev)
        prob_if_out = normal.cdf((clean_values - threshold_if_out)/noise_stddev)
        prob = torch.where(is_in, prob_if_in, prob_if_out)
        return prob

    def noisy_top_k_gating(self, x, train, noise_epsilon=1e-2):
        """Noisy top-k gating.
        See paper: <https://arxiv.org/abs/1701.06538>.
        Args:
            x: input Tensor with shape [batch_size, input_size]
            train: a boolean - we only add noise at training time.
            noise_epsilon: a float
        Returns:
            gates: a Tensor with shape [batch_size, num_experts]
            load: a Tensor with shape [num_experts]
        """
        clean_logits = x @ self.w_gate
        if self.noisy_gating and train:
            raw_noise_stddev = x @ self.w_noise
            noise_stddev = self.softplus(raw_noise_stddev) + noise_epsilon
            noisy_logits = clean_logits + (torch.randn_like(clean_logits) * noise_stddev)
            logits = noisy_logits
        else:
            logits = clean_logits

        # calculate topk + 1 that will be needed for the noisy gates
        logits = self.softmax(logits)
        top_logits, top_indices = logits.topk(min(self.k + 1, self.num_experts), dim=-1)
        top_k_logits = top_logits[:, :self.k]
        top_k_indices = top_indices[:, :self.k]
        top_k_gates = top_k_logits / (top_k_logits.sum(1, keepdim=True) + 1e-6)  # normalization

        zeros = torch.zeros_like(logits, requires_grad=True)
        gates = zeros.scatter(-1, top_k_indices, top_k_gates)  # non-topk elements will be 0

        if self.noisy_gating and self.k < self.num_experts and train:
            load = (self._prob_in_top_k(clean_logits, noisy_logits, noise_stddev, top_logits)).sum(0)
        else:
            load = self._gates_to_load(gates)
        return gates, load

    def forward(self, x, loss_coef=0.):
        """
        Token/Node-level Gating with the default gating algorithm in <https://arxiv.org/abs/1701.06538>.
        In specific, each token/node chooses TopK experts, auxiliary losses required for load balancing.
        Empirically, we found that the load-balancing loss may conflict with the reinforcement loss,
        and thus we do not return loss (i.e., loss_coef=0.).
        Please Refer to (Zhou et al, 2024) <https://arxiv.org/abs/2405.01029>.

        Args:
            x: tensor shape [batch_size, problem_size, input_size]
            loss_coef: a scalar - multiplier on load-balancing losses

        Returns:
            y: a tensor with shape [batch_size, problem_size, output_size].
            loss: a scalar. This should be added into the overall training loss of the model.
                The backpropagation of this loss encourages all experts to be approximately equally used across a batch.
        """
        output_shape = list(x.size()[:-1]) + [self.output_size]
        x = x.reshape(-1, self.input_size) if x.dim() != 2 else x

        gates, load = self.noisy_top_k_gating(x, self.training)
        # calculate importance loss
        importance = gates.sum(0)
        loss = self.cv_squared(importance) + self.cv_squared(load)
        loss *= loss_coef

        dispatcher = SparseDispatcher(self.num_experts, gates)
        expert_inputs = dispatcher.dispatch(x)
        # gates = dispatcher.expert_to_gates()
        expert_outputs = [self.experts[i](expert_inputs[i]) for i in range(self.num_experts)]
        y = dispatcher.combine(expert_outputs)

        # return y.reshape(output_shape), loss
        return y.reshape(output_shape)



from typing import Optional

import torch
import torch.nn.functional as F

from tensordict.tensordict import TensorDict
from torchrl.data import Bounded, Composite, Unbounded

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.utils.ops import gather_by_index, get_tour_length
from rl4co.utils.pylogger import get_pylogger

from .generator import OPGenerator
from .render import render

log = get_pylogger(__name__)


class OPEnv(RL4COEnvBase):
    """Orienteering Problem (OP) environment.
    At each step, the agent chooses a location to visit in order to maximize the collected prize.
    The total length of the path must not exceed a given threshold.

    Observations:
        - location of the depot
        - locations and prize of each customer
        - current location of the vehicle
        - current tour length
        - current total prize
        - the remaining length of the path

    Constraints:
        - the tour starts and ends at the depot
        - not all customers need to be visited
        - the vehicle cannot visit customers exceed the remaining length of the path

    Finish Condition:
        - the vehicle back to the depot

    Reward:
        - the sum of the prizes of visited nodes

    Args:
        generator: OPGenerator instance as the data generator
        generator_params: parameters for the generator
    """

    name = "op"

    def __init__(
        self,
        generator: OPGenerator = None,
        generator_params: dict = {},
        prize_type: str = "dist",
        **kwargs,
    ):
        super().__init__(**kwargs)
        if generator is None:
            generator = OPGenerator(**generator_params)
        self.generator = generator
        self.prize_type = prize_type
        assert self.prize_type in [
            "dist",
            "unif",
            "const",
        ], f"Invalid prize_type: {self.prize_type}"
        self._make_spec(self.generator)

    def _step(self, td: TensorDict) -> TensorDict:
        current_node = td["action"][:, None]

        # Update tour length
        previus_loc = gather_by_index(td["locs"], td["current_node"])
        current_loc = gather_by_index(td["locs"], current_node)
        tour_length = td["tour_length"] + (current_loc - previus_loc).norm(p=2, dim=-1)

        # Update prize with collected prize
        current_total_prize = td["current_total_prize"] + gather_by_index(
            td["prize"], current_node, dim=-1
        )

        # Set current node as visited
        visited = td["visited"].scatter(-1, current_node, 1)

        # Done if went back to depot (except if it's the first step, since we start at the depot)
        done = (current_node.squeeze(-1) == 0) & (td["i"] > 0)

        # The reward is calculated outside via get_reward for efficiency, so we set it to 0 here
        reward = torch.zeros_like(done)

        td.update(
            {
                "tour_length": tour_length,
                "current_node": current_node,
                "visited": visited,
                "current_total_prize": current_total_prize,
                "i": td["i"] + 1,
                "reward": reward,
                "done": done,
            }
        )
        td.set("action_mask", self.get_action_mask(td))
        return td

    def _reset(
        self,
        td: Optional[TensorDict] = None,
        batch_size: Optional[list] = None,
    ) -> TensorDict:
        device = td.device

        # Add depot to locs
        locs_with_depot = torch.cat((td["depot"][:, None, :], td["locs"]), -2)

        # Create reset TensorDict
        td_reset = TensorDict(
            {
                "locs": locs_with_depot,
                "prize": F.pad(
                    td["prize"], (1, 0), mode="constant", value=0
                ),  # add 0 for depot
                "tour_length": torch.zeros(*batch_size, device=device),
                # max_length is max length allowed when arriving at node, so subtract distance to return to depot
                # Additionally, substract epsilon margin for numeric stability
                "max_length": td["max_length"][..., None]
                - (td["depot"][..., None, :] - locs_with_depot).norm(p=2, dim=-1)
                - 1e-6,
                "current_node": torch.zeros(
                    *batch_size, 1, dtype=torch.long, device=device
                ),
                "visited": torch.zeros(
                    (*batch_size, locs_with_depot.shape[-2]),
                    dtype=torch.bool,
                    device=device,
                ),
                "current_total_prize": torch.zeros(
                    *batch_size, dtype=torch.float, device=device
                ),
                "i": torch.zeros(
                    (*batch_size,), dtype=torch.int64, device=device
                ),  # counter
            },
            batch_size=batch_size,
        )
        td_reset.set("action_mask", self.get_action_mask(td_reset))
        return td_reset

    @staticmethod
    def get_action_mask(td: TensorDict) -> torch.Tensor:
        """Get action mask with 1 = feasible action, 0 = infeasible action.
        Cannot visit if already visited, if depot has been visited, or if the length exceeds the maximum length.
        """
        current_loc = gather_by_index(td["locs"], td["current_node"])[..., None, :]
        exceeds_length = (
            td["tour_length"][..., None] + (td["locs"] - current_loc).norm(p=2, dim=-1)
            > td["max_length"]
        )
        mask = td["visited"] | td["visited"][..., 0:1] | exceeds_length

        action_mask = ~mask  # 1 = feasible action, 0 = infeasible action

        # Depot can always be visited: we do not hardcode knowledge that this is strictly suboptimal if other options are available
        action_mask[..., 0] = 1
        return action_mask

    def _get_reward(self, td: TensorDict, actions: torch.Tensor) -> torch.Tensor:
        """Reward is the sum of the prizes of visited nodes"""
        # In case all tours directly return to depot, prevent further problems
        if actions.size(-1) == 1:
            assert (actions == 0).all(), "If all length 1 tours, they should be zero"
            return torch.zeros(actions.size(0), dtype=torch.float, device=actions.device)

        # Prize is the sum of the prizes of the visited nodes. Note that prize is padded with 0 for depot at index 0
        collected_prize = td["prize"].gather(1, actions)
        return collected_prize.sum(-1)

    @staticmethod
    def check_solution_validity(
        td: TensorDict, actions: torch.Tensor, add_distance_to_depot: bool = True
    ) -> None:
        """Check that solution is valid: nodes are not visited twice except depot and capacity is not exceeded.
        If `add_distance_to_depot` if True, then the distance to the depot is added to max length since by default, the max length is
        modified in the reset function to account for the distance to the depot.
        """

        # Check that tours are valid, i.e. contain 0 to n -1
        sorted_actions = actions.data.sort(1)[0]
        # Make sure each node visited once at most (except for depot)
        assert (
            (sorted_actions[:, 1:] == 0)
            | (sorted_actions[:, 1:] > sorted_actions[:, :-1])
        ).all(), "Duplicates"

        # Gather locations in order of tour and get the length of tours
        locs_ordered = gather_by_index(td["locs"], actions)
        length = get_tour_length(locs_ordered)

        max_length = td["max_length"]
        if add_distance_to_depot:
            max_length = (
                max_length
                + (td["locs"][..., 0:1, :] - td["locs"]).norm(p=2, dim=-1)
                + 1e-6
            )
        assert (
            length[..., None] <= max_length + 1e-5
        ).all(), "Max length exceeded by {}".format(
            (length[..., None] - max_length).max()
        )

    def _make_spec(self, generator: OPGenerator):
        """Make the observation and action specs from the parameters."""
        self.observation_spec = Composite(
            locs=Bounded(
                low=generator.min_loc,
                high=generator.max_loc,
                shape=(generator.num_loc + 1, 2),
                dtype=torch.float32,
            ),
            current_node=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            prize=Unbounded(
                shape=(generator.num_loc,),
                dtype=torch.float32,
            ),
            tour_length=Unbounded(
                shape=(generator.num_loc,),
                dtype=torch.float32,
            ),
            visited=Unbounded(
                shape=(generator.num_loc + 1,),
                dtype=torch.bool,
            ),
            max_length=Unbounded(
                shape=(1,),
                dtype=torch.float32,
            ),
            action_mask=Unbounded(
                shape=(generator.num_loc + 1, 1),
                dtype=torch.bool,
            ),
            shape=(),
        )
        self.action_spec = Bounded(
            shape=(1,),
            dtype=torch.int64,
            low=0,
            high=generator.num_loc + 1,
        )
        self.reward_spec = Unbounded(shape=(1,))
        self.done_spec = Unbounded(shape=(1,), dtype=torch.bool)

    @staticmethod
    def render(td: TensorDict, actions: torch.Tensor = None, ax=None):
        return render(td, actions, ax)



import abc

from typing import Any, Callable, Optional, Tuple

import torch.nn as nn

from tensordict import TensorDict
from torch import Tensor

from rl4co.envs import RL4COEnvBase, get_env
from rl4co.utils.decoding import (
    DecodingStrategy,
    get_decoding_strategy,
    get_log_likelihood,
)
from rl4co.utils.ops import calculate_entropy
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class ConstructiveEncoder(nn.Module, metaclass=abc.ABCMeta):
    """Base class for the encoder of constructive models"""

    @abc.abstractmethod
    def forward(self, td: TensorDict) -> Tuple[Any, Tensor]:
        """Forward pass for the encoder

        Args:
            td: TensorDict containing the input data

        Returns:
            Tuple containing:
              - latent representation (any type)
              - initial embeddings (from feature space to embedding space)
        """
        raise NotImplementedError("Implement me in subclass!")


class ConstructiveDecoder(nn.Module, metaclass=abc.ABCMeta):
    """Base decoder model for constructive models. The decoder is responsible for generating the logits for the action"""

    @abc.abstractmethod
    def forward(
        self, td: TensorDict, hidden: Any = None, num_starts: int = 0
    ) -> Tuple[Tensor, Tensor]:
        """Obtain logits for current action to the next ones

        Args:
            td: TensorDict containing the input data
            hidden: Hidden state from the encoder. Can be any type
            num_starts: Number of starts for multistart decoding

        Returns:
            Tuple containing the logits and the action mask
        """
        raise NotImplementedError("Implement me in subclass!")

    def pre_decoder_hook(
        self, td: TensorDict, env: RL4COEnvBase, hidden: Any = None, num_starts: int = 0
    ) -> Tuple[TensorDict, RL4COEnvBase, Any]:
        """By default, we don't need to do anything here.

        Args:
            td: TensorDict containing the input data
            hidden: Hidden state from the encoder
            env: Environment for decoding
            num_starts: Number of starts for multistart decoding

        Returns:
            Tuple containing the updated Tensordict, environment, and hidden state
        """
        return td, env, hidden


class NoEncoder(ConstructiveEncoder):
    """Default encoder decoder-only models, i.e. autoregressive models that re-encode all the state at each decoding step."""

    def forward(self, td: TensorDict) -> Tuple[Tensor, Tensor]:
        """Return Nones for the hidden state and initial embeddings"""
        return None, None


class ConstructivePolicy(nn.Module):
    """
    Base class for constructive policies. Constructive policies take as input and instance and output a solution (sequence of actions).
    "Constructive" means that a solution is created from scratch by the model.

    The structure follows roughly the following steps:
        1. Create a hidden state from the encoder
        2. Initialize decoding strategy (such as greedy, sampling, etc.)
        3. Decode the action given the hidden state and the environment state at the current step
        4. Update the environment state with the action. Repeat 3-4 until all sequences are done
        5. Obtain log likelihood, rewards etc.

    Note that an encoder is not strictly needed (see :class:`NoEncoder`).). A decoder however is always needed either in the form of a
    network or a function.

    Note:
        There are major differences between this decoding and most RL problems. The most important one is
        that reward may not defined for partial solutions, hence we have to wait for the environment to reach a terminal
        state before we can compute the reward with `env.get_reward()`.

    Warning:
        We suppose environments in the `done` state are still available for sampling. This is because in NCO we need to
        wait for all the environments to reach a terminal state before we can stop the decoding process. This is in
        contrast with the TorchRL framework (at the moment) where the `env.rollout` function automatically resets.
        You may follow tighter integration with TorchRL here: https://github.com/ai4co/rl4co/issues/72.

    Args:
        encoder: Encoder to use
        decoder: Decoder to use
        env_name: Environment name to solve (used for automatically instantiating networks)
        temperature: Temperature for the softmax during decoding
        tanh_clipping: Clipping value for the tanh activation (see Bello et al. 2016) during decoding
        mask_logits: Whether to mask the logits or not during decoding
        train_decode_type: Decoding strategy for training
        val_decode_type: Decoding strategy for validation
        test_decode_type: Decoding strategy for testing
    """

    def __init__(
        self,
        encoder: ConstructiveEncoder | Callable,
        decoder: ConstructiveDecoder | Callable,
        env_name: str = "tsp",
        temperature: float = 1.0,
        tanh_clipping: float = 0,
        mask_logits: bool = True,
        train_decode_type: str = "sampling",
        val_decode_type: str = "greedy",
        test_decode_type: str = "greedy",
        **unused_kw,
    ):
        super(ConstructivePolicy, self).__init__()

        if len(unused_kw) > 0:
            log.error(f"Found {len(unused_kw)} unused kwargs: {unused_kw}")

        self.env_name = env_name

        # Encoder and decoder
        if encoder is None:
            log.warning("`None` was provided as encoder. Using `NoEncoder`.")
            encoder = NoEncoder()
        self.encoder = encoder
        self.decoder = decoder

        # Decoding strategies
        self.temperature = temperature
        self.tanh_clipping = tanh_clipping
        self.mask_logits = mask_logits
        self.train_decode_type = train_decode_type
        self.val_decode_type = val_decode_type
        self.test_decode_type = test_decode_type

    def forward(
        self,
        td: TensorDict,
        env: Optional[str | RL4COEnvBase] = None,
        phase: str = "train",
        calc_reward: bool = True,
        return_actions: bool = True,
        return_entropy: bool = False,
        return_hidden: bool = False,
        return_init_embeds: bool = False,
        return_sum_log_likelihood: bool = True,
        actions=None,
        max_steps=1_000_000,
        **decoding_kwargs,
    ) -> dict:
        """Forward pass of the policy.

        Args:
            td: TensorDict containing the environment state
            env: Environment to use for decoding. If None, the environment is instantiated from `env_name`. Note that
                it is more efficient to pass an already instantiated environment each time for fine-grained control
            phase: Phase of the algorithm (train, val, test)
            calc_reward: Whether to calculate the reward
            return_actions: Whether to return the actions
            return_entropy: Whether to return the entropy
            return_hidden: Whether to return the hidden state
            return_init_embeds: Whether to return the initial embeddings
            return_sum_log_likelihood: Whether to return the sum of the log likelihood
            actions: Actions to use for evaluating the policy.
                If passed, use these actions instead of sampling from the policy to calculate log likelihood
            max_steps: Maximum number of decoding steps for sanity check to avoid infinite loops if envs are buggy (i.e. do not reach `done`)
            decoding_kwargs: Keyword arguments for the decoding strategy. See :class:`rl4co.utils.decoding.DecodingStrategy` for more information.

        Returns:
            out: Dictionary containing the reward, log likelihood, and optionally the actions and entropy
        """

        # Encoder: get encoder output and initial embeddings from initial state
        hidden, init_embeds = self.encoder(td)

        # Instantiate environment if needed
        if isinstance(env, str) or env is None:
            env_name = self.env_name if env is None else env
            log.info(f"Instantiated environment not provided; instantiating {env_name}")
            env = get_env(env_name)

        # Get decode type depending on phase and whether actions are passed for evaluation
        decode_type = decoding_kwargs.pop("decode_type", None)
        if actions is not None:
            decode_type = "evaluate"
        elif decode_type is None:
            decode_type = getattr(self, f"{phase}_decode_type")

        # Setup decoding strategy
        # we pop arguments that are not part of the decoding strategy
        decode_strategy: DecodingStrategy = get_decoding_strategy(
            decode_type,
            temperature=decoding_kwargs.pop("temperature", self.temperature),
            tanh_clipping=decoding_kwargs.pop("tanh_clipping", self.tanh_clipping),
            mask_logits=decoding_kwargs.pop("mask_logits", self.mask_logits),
            store_all_logp=decoding_kwargs.pop("store_all_logp", return_entropy),
            **decoding_kwargs,
        )

        # Pre-decoding hook: used for the initial step(s) of the decoding strategy
        td, env, num_starts = decode_strategy.pre_decoder_hook(td, env)

        # Additionally call a decoder hook if needed before main decoding
        td, env, hidden = self.decoder.pre_decoder_hook(td, env, hidden, num_starts)

        # Main decoding: loop until all sequences are done
        step = 0
        while not td["done"].all():
            logits, mask = self.decoder(td, hidden, num_starts)
            td = decode_strategy.step(
                logits,
                mask,
                td,
                action=actions[..., step] if actions is not None else None,
            )
            td = env.step(td)["next"]
            step += 1
            if step > max_steps:
                log.error(
                    f"Exceeded maximum number of steps ({max_steps}) duing decoding"
                )
                break

        # Post-decoding hook: used for the final step(s) of the decoding strategy
        logprobs, actions, td, env = decode_strategy.post_decoder_hook(td, env)

        # Output dictionary construction
        if calc_reward:
            td.set("reward", env.get_reward(td, actions))

        outdict = {
            "reward": td["reward"],
            "log_likelihood": get_log_likelihood(
                logprobs, actions, td.get("mask", None), return_sum_log_likelihood
            ),
        }

        if return_actions:
            outdict["actions"] = actions
        if return_entropy:
            outdict["entropy"] = calculate_entropy(logprobs)
        if return_hidden:
            outdict["hidden"] = hidden
        if return_init_embeds:
            outdict["init_embeds"] = init_embeds

        return outdict
